import re, random
from typing import Optional, Dict, Any, List, Tuple, Union
from textarena.envs.Coup.coup_types import CoupActionType, GamePhase, ActionMetadata

import textarena as ta
from textarena.envs.Coup import base_coup_prompts

from rich.text import Text




class CoupEnv(ta.Env):
    """
    A minimal text-based implementation of the Coup card game using textarena.

    Example actions a player might enter:
      - [<action> <target_player_id> ]
      - [PASS]                 -> The player makes an action.
      - [BULLSHIT]             -> The player challenges the last play.
      - [block foreign aid]    -> The player counteracts the last play.
    """

    

    def __init__(self):
        super().__init__()

    def reset(self, num_players: int, seed: Optional[int] = None) -> None:
        """ Reset the environment for a new Coup game """
        self.state = ta.State(num_players=num_players, min_players=2, max_players=6)
        self.state.error_allowance = 3

        # Create deck with three of each card
        deck = ["Duke", "Assassin", "Ambassador", "Captain", "Contessa"] * 3
        random.shuffle(deck)

        # Initialize game state
        game_state = {
            "phase": GamePhase.Play,

            "hidden_hand": {},  # The cards each player has in their hand
            "revealed_hand": {},  # The cards each player has revealed/lost
            "pile": [],  # The cards in the pile in the middle
            
            "coins": {pid: 2 for pid in range(num_players)},  # Each player starts with 2 coins
            "treasury_coins": 50 - 2*num_players, # Max 50 coins in the pot, minus the coins each player starts with

            "action_metadata": None, # The metadata about the current action that is being taken (can span multiple "turns")
        }

        # Deal two cards to each player
        for player_id in range(num_players):
            game_state["hidden_hand"].setdefault(player_id, [])
            game_state["revealed_hand"].setdefault(player_id, [])
            game_state["hidden_hand"][player_id].append(deck.pop())
            game_state["hidden_hand"][player_id].append(deck.pop())

        # Pile has whatever is left in the deck
        game_state["pile"] = deck
        
        # Add the rendered board to the game state
        game_state["rendered_board"] = self._render_board(game_state)
        
        # Initialize textarena state
        self.state.reset(seed=seed, game_state=game_state, player_prompt_function=self._gen_initial_prompt)

        # Always start with player 0
        self.state.manually_update_current_player(new_player_id=0)
        self._send_call_to_action_prompt()
            
    
    def get_board_str(self):
        return self.state.game_state["rendered_board"]

    def step(self, action: str) -> Tuple[bool, ta.Info]:
        """ Process a single step/action from the current player """
        # Log the player's raw input by sending the action_str to player with id -1
        self.state.add_observation(from_id=self.state.current_player_id, to_id=-1, message=action, for_logging=True)

        # Start with the assumption that we can move to the next player. (is set to True if we detect an invalid move)
        self.state.prevent_player_change = False

        # Game phase is either in play (a player is doing an initial action like income, foreign aid, etc)
        # or challenge (we are querying one or more players about a potential counteraction)
        try:

            action_type, other_stuff = self._parse_action(action)
            if self.state.game_state["phase"] == GamePhase.Play:
                self._update_action_metadata_for_play_phase(action_type, other_stuff)  # other stuff here is the target player id
            elif self.state.game_state["phase"] == GamePhase.QueryForBlockOrChallenge:
                self._update_action_metadata_for_query_for_block_or_challenge_phase(action_type)
            elif self.state.game_state["phase"] == GamePhase.QueryToChallengeTheBlocker:
                self._update_action_metadata_for_query_to_challenge_the_blocker_phase(action_type)
            elif self.state.game_state["phase"] == GamePhase.QueryWhichToKeep:
                self._update_action_metadata_for_query_which_to_keep_phase(action_type, other_stuff)
            else:
                raise Exception(f"Unexpected game phase: {self.state.game_state['phase']}")

            
            if self.state.prevent_player_change:
                return False, {"reason": "Invalid move"}
            
            adv_turn = self._advance_turn()

            # Render after we've updated the action metadata and advanced the turn
            self.state.game_state["rendered_board"] = self._render_board()
            
            # This tells the current player it's their turn and asks them what they want to do.
            if not adv_turn[0]:  # Game not over
                self._send_call_to_action_prompt()

            return adv_turn
        except ValueError as e:
            self.state.set_invalid_move(player_id=self.state.current_player_id, reason=str(e))
            return False, {"reason":str(e)}

        
    def _update_action_metadata_for_play_phase(self, action_type: CoupActionType, action_target_player_id: Optional[int] = None):
        """
        This validates state and updates state as needed. NOTE THAT ONLY ALLOWABLE ACTIONS HERE ARE THE ONES FROM THE OFFICIAL CHEATSHEET "ACTION" COLUMN (see README.md)
        """
        if self.state.game_state["coins"][self.state.current_player_id] >= 10 and action_type is not CoupActionType.Coup:
            self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. You cannot do anything other than coup when you have 10 or more coins. Please pick a player id to Coup and respond with: [coup x].")
            return
        if action_type is CoupActionType.Income:
            self.state.game_state["action_metadata"] = ActionMetadata(action_type=action_type, source_player_id=self.state.current_player_id, target_player_id=action_target_player_id)
            self._execute_current_action()
        elif action_type is CoupActionType.Coup:
            if self.state.game_state["coins"][self.state.current_player_id] < 7:
                self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. You don't have enough coins to coup.")
                return
            if self.state.game_state["hidden_hand"][action_target_player_id] == []:
                self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. Can't coup player {action_target_player_id} because they are already eliminated from play.")
                return
            
            self.state.game_state["action_metadata"] = ActionMetadata(action_type=action_type, source_player_id=self.state.current_player_id, target_player_id=action_target_player_id)
            self._execute_current_action()

        elif action_type is CoupActionType.Exchange or action_type is CoupActionType.Tax:  # Exchange and Tax are not blockable, so we just skip to QueryForBlockOrChallenge phase
            self.state.game_state["phase"] = GamePhase.QueryForBlockOrChallenge
            active_players = [pid for pid in range(self.state.num_players) if len(self.state.game_state["hidden_hand"][pid]) > 0 and pid != self.state.current_player_id]
            self.state.game_state["action_metadata"] = ActionMetadata(action_type=action_type, source_player_id=self.state.current_player_id, players_to_query=active_players)
        
        elif action_type is CoupActionType.ForeignAid or action_type is CoupActionType.Assassinate or action_type is CoupActionType.Steal:
            if action_type is CoupActionType.Assassinate and self.state.game_state["coins"][self.state.current_player_id] < 3:
                self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. You don't have enough coins to assassinate.")
                return
            if action_type is CoupActionType.Steal and action_target_player_id is not None and action_target_player_id == self.state.current_player_id:
                self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. You cannot steal from yourself, the second value in the [steal x] response must be a valid player id.")
                return
            if action_type is CoupActionType.Steal and action_target_player_id is None and action_target_player_id >= self.state.num_players:
                self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. The second value in the [steal x] response must be a valid player id.")
                return
            if action_type is CoupActionType.Steal and self.state.game_state["coins"][action_target_player_id] < 2:
                self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. Player {action_target_player_id} doesn't have enough coins to steal from.")
                return
            if action_type is CoupActionType.Steal and self.state.game_state["hidden_hand"][action_target_player_id] == []:
                self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. You cannot steal from Player {action_target_player_id} because they are already eliminated from play.")
                return

            # For assassination, deduct the cost immediately (per rules, you pay even if blocked)
            if action_type is CoupActionType.Assassinate:
                self.state.game_state["coins"][self.state.current_player_id] -= 3
                self.state.game_state["treasury_coins"] += 3

            self.state.game_state["phase"] = GamePhase.QueryForBlockOrChallenge
            # Only query active players (those with cards remaining)
            active_players = [pid for pid in range(self.state.num_players) if len(self.state.game_state["hidden_hand"][pid]) > 0 and pid != self.state.current_player_id]
            # For targeted actions, ensure target is first in query list if they're active
            if action_target_player_id is not None and action_target_player_id in active_players:
                active_players.remove(action_target_player_id)
                active_players = [action_target_player_id] + active_players
            self.state.game_state["action_metadata"] = ActionMetadata(action_type=action_type, source_player_id=self.state.current_player_id, target_player_id=action_target_player_id, players_to_query=active_players)

    def _update_action_metadata_for_query_for_block_or_challenge_phase(self, action: CoupActionType):
        # If the player passes (they don't want to challenge), then we don't need to do anything, just advance (this is handled in step())
        if action is CoupActionType.PASS:
            return
        elif action is CoupActionType.BULLSHIT:
            self.state.game_state["action_metadata"].challenger_player_id = self.state.current_player_id
            self._execute_showdown_on_bullshit()
        elif action is CoupActionType.BlockForeignAid or action is CoupActionType.BlockStealAmbassador or action is CoupActionType.BlockStealCaptain or action is CoupActionType.BlockAssassinate:
            if (action is CoupActionType.BlockStealAmbassador or action is CoupActionType.BlockStealCaptain) and self.state.game_state["action_metadata"].action_type is not CoupActionType.Steal:
                self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. You cannot call [block steal x] when the last played action is a {self.state.game_state['action_metadata'].action_type.name}.")
                return
            if action is CoupActionType.BlockAssassinate and self.state.game_state["action_metadata"].action_type is not CoupActionType.Assassinate:
                self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. You cannot call [block assassinate] when the last played action is a {self.state.game_state['action_metadata'].action_type.name}.")
                return
            if action is CoupActionType.BlockForeignAid and self.state.game_state["action_metadata"].action_type is not CoupActionType.ForeignAid:
                self.state.set_invalid_move(player_id=self.state.current_player_id, reason=f"Invalid move. You cannot call [block foreign aid] when the last played action is a {self.state.game_state['action_metadata'].action_type.name}.")
                return
            self.state.game_state["phase"] = GamePhase.QueryToChallengeTheBlocker
            self.state.game_state["action_metadata"].blocker_player_id = self.state.current_player_id
            self.state.game_state["action_metadata"].block_type = action  # Track which specific block was used

            # This is a bit of a hack to ensure that the source player is always first in the query list.
            active_players = [self.state.game_state["action_metadata"].source_player_id] + [pid for pid in range(self.state.num_players) if len(self.state.game_state["hidden_hand"][pid]) > 0 and pid != self.state.current_player_id and pid != self.state.game_state["action_metadata"].source_player_id]
            self.state.game_state["action_metadata"].players_to_query = active_players
        else:  # Anything other than PASS or BULLSHIT or BlockXXXX is invalid in this phase
            block_options_str = "[block steal captain], [block steal ambassador], " if self.state.game_state["action_metadata"].action_type is CoupActionType.Steal and self.state.game_state["action_metadata"].target_player_id == self.state.current_player_id else \
                ("[block assassinate], " if (self.state.game_state["action_metadata"].action_type is CoupActionType.Assassinate  and self.state.game_state["action_metadata"].target_player_id == self.state.current_player_id) else \
                ("[block foreign aid], " if self.state.game_state["action_metadata"].action_type is CoupActionType.ForeignAid else ""))
            bullshit_str = "call [BULLSHIT], " if self.state.game_state["action_metadata"].action_type is not CoupActionType.ForeignAid else "" # you can't bullshit on foreign aid
            raise ValueError(f"Invalid action: {action}, Player #{self.state.game_state['action_metadata'].source_player_id} is attempting to {self.state.game_state["action_metadata"].action_type}, you must either {block_options_str}{bullshit_str}or [PASS]")

    def _update_action_metadata_for_query_to_challenge_the_blocker_phase(self, action: CoupActionType):
        if action is CoupActionType.PASS:
            return
        elif action is CoupActionType.BULLSHIT:
            self.state.game_state["action_metadata"].blocker_challenger_player_id = self.state.current_player_id
            self._execute_showdown_on_blocker_bullshit()
        else:
            raise ValueError(f"Invalid action: {action}, you must either [PASS] or call [BULLSHIT]")

    def _update_action_metadata_for_query_which_to_keep_phase(self, action_type: CoupActionType, cards_to_keep: Optional[List[str]] = None):
        """
        Handle the phase where player is choosing which cards to keep after an exchange
        """
        if action_type is not CoupActionType.Keep:
            cards_str = "<card1> <card2>" if len(self.state.game_state["revealed_hand"][self.state.current_player_id]) == 0 else "<card>"
            raise ValueError(f"Invalid action: {action_type}, you must respond with [keep {cards_str}] to specify the card(s) to keep")
        
        if cards_to_keep is None:
            raise ValueError("You must specify which cards to keep")
            
        # Check how many cards the player should keep based on their remaining influences
        player_id = self.state.current_player_id
        expected_cards_to_keep = 2 - len(self.state.game_state["revealed_hand"][player_id])
        
        if len(cards_to_keep) != expected_cards_to_keep:
            if expected_cards_to_keep == 2:
                raise ValueError("You must specify exactly two cards to keep")
            else:
                raise ValueError("You must specify exactly one card to keep (since you have lost an influence)")
        
        # Store the cards to keep in metadata
        self.state.game_state["action_metadata"].cards_to_keep = [c.title() for c in cards_to_keep]
        
        # Complete the exchange action
        self._execute_exchange_action()

    ######################################################################################################################################################################
    # These are all the "sink" states for a single turn. Either we execute the action, or we resolve a block or challenge and then potentially execute the action still. #
    ######################################################################################################################################################################
    def _execute_current_action(self):
        curr_action = self.state.game_state["action_metadata"]
        
        # This means no-op, a challenge/block was successful and we're not executing the action
        if curr_action.action_type is CoupActionType.PASS:
            return
        
        source_player_observation, target_player_observation, other_player_observations = None, None, None
        
        if curr_action.action_type is CoupActionType.Income:
            self.state.game_state["coins"][curr_action.source_player_id] += 1
            self.state.game_state["treasury_coins"] -= 1
            source_player_observation = f"You just successfully played income. You now have {self.state.game_state['coins'][curr_action.source_player_id]} coins"
            other_player_observations = f"Player #{curr_action.source_player_id} just played income. They now have {self.state.game_state['coins'][curr_action.source_player_id]} coins"
            
        elif curr_action.action_type is CoupActionType.Coup:
            self.state.game_state["coins"][curr_action.source_player_id] -= 7
            self.state.game_state["treasury_coins"] += 7

            # Make the target player lose a card
            lost_card = self._make_player_lose_a_card(curr_action.target_player_id)

            remaining_cards_with_target = "now have 1 card remaining." if len(self.state.game_state["hidden_hand"][curr_action.target_player_id]) > 0 else "now have no cards remaining and are eliminated from play."
            source_player_observation = f"You just successfully played coup on Player #{curr_action.target_player_id}. They lost a card and revealed a {lost_card}. They {remaining_cards_with_target}"
            target_player_observation = f"You just lost a card and revealed a {lost_card} card. You {remaining_cards_with_target}"
            other_player_observations = f"Player #{curr_action.source_player_id} just played coup on Player #{curr_action.target_player_id}. They lost a card and revealed a {lost_card} card. They {remaining_cards_with_target}"
            
        elif curr_action.action_type is CoupActionType.ForeignAid:
            self.state.game_state["coins"][curr_action.source_player_id] += 2
            self.state.game_state["treasury_coins"] -= 2

            source_player_observation = f"You just successfully played foreign aid. You now have {self.state.game_state['coins'][curr_action.source_player_id]} coins"
            other_player_observations = f"Player #{curr_action.source_player_id} just played foreign aid. They now have {self.state.game_state['coins'][curr_action.source_player_id]} coins"

        elif curr_action.action_type is CoupActionType.Tax:
            self.state.game_state["coins"][curr_action.source_player_id] += 3
            self.state.game_state["treasury_coins"] -= 3

            source_player_observation = f"You just successfully played tax. You now have {self.state.game_state['coins'][curr_action.source_player_id]} coins"
            other_player_observations = f"Player #{curr_action.source_player_id} just played tax. They now have {self.state.game_state['coins'][curr_action.source_player_id]} coins"

        elif curr_action.action_type is CoupActionType.Assassinate:
            # Check for if player is already out, do nothing if so.
            if len(self.state.game_state["hidden_hand"][curr_action.target_player_id]) == 0:
                return
            
            # Make the target player lose a card
            lost_card = self._make_player_lose_a_card(curr_action.target_player_id)
            remaining_cards_with_target = "now have 1 card remaining." if len(self.state.game_state["hidden_hand"][curr_action.target_player_id]) > 0 else "now have no cards remaining and are eliminated from play."
            source_player_observation = f"You just successfully played assassinate on Player #{curr_action.target_player_id}. They lost a card and revealed a {lost_card}. They {remaining_cards_with_target}"
            target_player_observation = f"You just lost a card and revealed a {lost_card} card. You {remaining_cards_with_target}"
            other_player_observations = f"Player #{curr_action.source_player_id} just played assassinate on Player #{curr_action.target_player_id}. They lost a card and revealed a {lost_card}. They {remaining_cards_with_target}"
            
        elif curr_action.action_type is CoupActionType.Steal:
            self.state.game_state["coins"][curr_action.source_player_id] += 2
            self.state.game_state["coins"][curr_action.target_player_id] -= 2

            source_player_observation = f"You just successfully stole two coins from Player #{curr_action.target_player_id}. You now have {self.state.game_state['coins'][curr_action.source_player_id]} coins. Player #{curr_action.target_player_id} has {self.state.game_state['coins'][curr_action.target_player_id]} coins"
            target_player_observation = f"Player #{curr_action.source_player_id} just stole two coins from you. You now have {self.state.game_state['coins'][curr_action.target_player_id]} coins. Player #{curr_action.source_player_id} has {self.state.game_state['coins'][curr_action.source_player_id]} coins"
            other_player_observations = f"Player #{curr_action.source_player_id} just stole two coins from Player #{curr_action.target_player_id}. Player #{curr_action.source_player_id} has {self.state.game_state['coins'][curr_action.source_player_id]} coins. Player #{curr_action.target_player_id} has {self.state.game_state['coins'][curr_action.target_player_id]} coins"
            
        elif curr_action.action_type is CoupActionType.Exchange:
            # Draw two cards from the pile
            if len(self.state.game_state["pile"]) < 2:
                raise ValueError("Not enough cards in pile for exchange")
            
            # Draw two cards and add them directly to the player's hand
            card1 = self.state.game_state["pile"].pop()
            card2 = self.state.game_state["pile"].pop()
            self.state.game_state["hidden_hand"][curr_action.source_player_id].append(card1)
            self.state.game_state["hidden_hand"][curr_action.source_player_id].append(card2)
            
            # Change phase to QueryWhichToKeep
            self.state.game_state["phase"] = GamePhase.QueryWhichToKeep
            
            # Send observation to player about their options
            all_cards = self.state.game_state["hidden_hand"][curr_action.source_player_id]
            cards_str = ", ".join(all_cards)
            
            # Determine how many cards they need to keep based on revealed cards (influences lost)
            cards_to_keep_count = 2 - len(self.state.game_state["revealed_hand"][curr_action.source_player_id])
            
            if cards_to_keep_count == 2:
                source_player_observation = f"You drew two cards from the pile for exchange. You now have: {cards_str}. You must choose which two cards to keep using [keep <card1>" + \
                    f"{' <card2>' if len(self.state.game_state['revealed_hand'][curr_action.source_player_id]) == 0 else ''}]"
            else:  # cards_to_keep_count == 1
                source_player_observation = f"You drew two cards from the pile for exchange. You now have: {cards_str}. Since you have lost an influence, you must choose which one card to keep using [keep <card>]"
                
            other_player_observations = f"Player #{curr_action.source_player_id} is exchanging cards with the Court deck."

        # Broadcast the observations to the players
        if source_player_observation:
            self.state.add_observation(from_id=ta.GAME_ID, to_id=curr_action.source_player_id, message=source_player_observation, for_logging=False)
        if target_player_observation is not None:
            self.state.add_observation(from_id=ta.GAME_ID, to_id=curr_action.target_player_id, message=target_player_observation, for_logging=False)
        if other_player_observations is not None:
            exclude_ids = [curr_action.source_player_id]
            if curr_action.target_player_id is not None:
                exclude_ids.append(curr_action.target_player_id)
            self._broadcast_observations(other_player_observations, exclude_player_ids=exclude_ids)


    def _execute_showdown_on_bullshit(self):
        # the "challenged" player is the player who initially played the challenged card
        challenger_player_id = self.state.game_state["action_metadata"].challenger_player_id
        challenged_card = self._action_to_card(self.state.game_state["action_metadata"].action_type)
        challenged_player_id = self.state.game_state["action_metadata"].source_player_id  # The person who made the original claim
        is_honest = self.state.game_state["hidden_hand"][challenged_player_id].count(challenged_card) > 0
        
        challenged_message = ""
        challenger_message = ""
        other_player_observations = ""
        
        if is_honest:
            # If player was honest, then he gets a new card from the pile, and we reshuffle. Then challenger loses a card
            self.state.game_state["hidden_hand"][challenged_player_id].remove(challenged_card)
            self.state.game_state["pile"].append(challenged_card)
            random.shuffle(self.state.game_state["pile"])

            new_pulled_card = self.state.game_state["pile"].pop()
            self.state.game_state["hidden_hand"][challenged_player_id].append(new_pulled_card)

            # Player who called bullshit loses a card
            card_lost_by_challenger = self._make_player_lose_a_card(challenger_player_id)
            
            
            ###### UPDATE OBSERVATIONS
            challenger_remaining_cards = self.state.game_state["hidden_hand"][challenger_player_id]

            # Check if the challenger was the target of assassination and is now eliminated
            if (self.state.game_state["action_metadata"].action_type is CoupActionType.Assassinate and 
                self.state.game_state["action_metadata"].target_player_id == challenger_player_id and
                len(challenger_remaining_cards) == 0):
                # Target is eliminated, so assassination should not proceed
                self.state.game_state["action_metadata"].action_type = CoupActionType.PASS
    
            # Tell the challenger that their bullshit call failed
            challenger_message = f"Your bullshit call on Player #{challenged_player_id} failed. They did indeed have a {challenged_card} card." + \
                (f"You lost a {card_lost_by_challenger} card." if (len(challenger_remaining_cards) == 0 or challenger_remaining_cards[0] != card_lost_by_challenger) else f"You lost one of your {card_lost_by_challenger} cards.") + \
                (f"You now have only one card remaining, the {challenger_remaining_cards[0]} card." if len(challenger_remaining_cards) > 0 else \
                f"You have no cards remaining, you're eliminated!")
            # Tell the challenged player that they just survived a bullshit challenge
            challenged_message = f"You were unsuccessfully challenged on your {challenged_card} claim by Player #{challenger_player_id}. " + \
                f"Because you had to reveal your {challenged_card} card to prove them wrong, you were given a new one from the pile. It is a {new_pulled_card} card. Player #{challenger_player_id} revealed and lost a {card_lost_by_challenger} card, " + \
                (f"they now have only one card remaining." if len(challenger_remaining_cards) > 0 else f"they have no cards remaining, Player #{challenger_player_id} is eliminated!")
            # Tell everyone else what happened
            other_player_observations = f"Player #{challenger_player_id} just unsuccessfully called bullshit on Player #{challenged_player_id}'s {challenged_card} claim! " + \
                f"Player #{challenged_player_id} did indeed have a {challenged_card} card, put it back in the pile and got a new one. Player #{challenger_player_id} lost an influence and revealed a {card_lost_by_challenger} card." + \
                (f"Player #{challenger_player_id} now has 1 card remaining." if len(challenger_remaining_cards) > 0 else f"Player #{challenger_player_id} has no cards remaining, they're eliminated!")
    
        else:
            # Player who got challenged loses a card, nothing else changes
            card_lost_by_challenged = self._make_player_lose_a_card(challenged_player_id)
            
            # If this was an assassination that got successfully challenged, refund the 3 coins
            if self.state.game_state["action_metadata"].action_type is CoupActionType.Assassinate:
                self.state.game_state["coins"][challenged_player_id] += 3
                self.state.game_state["treasury_coins"] -= 3
            
            challenged_message = f"You were challenged on your {challenged_card} claim by Player #{challenger_player_id}, Since you did not have a {challenged_card} card, you lost your {card_lost_by_challenged} card." + \
                (f" Your 3 coins for the assassination attempt have been refunded." if self.state.game_state["action_metadata"].action_type is CoupActionType.Assassinate else "") + \
                (f" You now have only one card remaining, the {self.state.game_state["hidden_hand"][challenged_player_id][0]} card." if len(self.state.game_state["hidden_hand"][challenged_player_id]) > 0 else \
                " You have no cards remaining, you're eliminated!")
            # Tell the challenger that they just successfully challenged the challenged player
            challenger_message = f"You just successfully challenged Player #{challenged_player_id} on their {challenged_card} claim! " + \
                (f"They are blocked from doing it and have {len(self.state.game_state["hidden_hand"][challenged_player_id])} card remaining." if len(self.state.game_state["hidden_hand"][challenged_player_id]) > 0 else \
                "They have no cards remaining, they're eliminated!") + \
                (f" They were refunded their 3 coins." if self.state.game_state["action_metadata"].action_type is CoupActionType.Assassinate else "")
            # Other players see the challenger successfully challenge the challenged player
            other_player_observations = f"Player #{challenger_player_id} just successfully challenged Player #{challenged_player_id} on their {challenged_card} claim! " + \
                (f"Player #{challenged_player_id} was blocked from doing it and has {len(self.state.game_state["hidden_hand"][challenged_player_id])} card remaining." if len(self.state.game_state["hidden_hand"][challenged_player_id]) > 0 else \
                f"Player #{challenged_player_id} has no cards remaining, they're eliminated!") + \
                (f" They were refunded their 3 coins." if self.state.game_state["action_metadata"].action_type is CoupActionType.Assassinate else "")
            
            # Mark no-op for _advance_turn()
            self.state.game_state["action_metadata"].action_type = CoupActionType.PASS

        # Also mark that we have no more players to query so that we can advance the turn
        self.state.game_state["action_metadata"].players_to_query = None
        self.state.add_observation(from_id=ta.GAME_ID, to_id=challenged_player_id, message=challenged_message, for_logging=False)
        self.state.add_observation(from_id=ta.GAME_ID, to_id=challenger_player_id, message=challenger_message, for_logging=False)
        self._broadcast_observations(other_player_observations, exclude_player_ids=[challenged_player_id, challenger_player_id])
    
    def _execute_showdown_on_blocker_bullshit(self):
        source_player_id = self.state.game_state["action_metadata"].source_player_id
        blocker_player_id = self.state.game_state["action_metadata"].blocker_player_id
        challenger_player_id = self.state.game_state["action_metadata"].blocker_challenger_player_id
        
        # Determine which card the blocker claimed to have based on the block type
        action_metadata = self.state.game_state["action_metadata"]
        block_type = getattr(action_metadata, 'block_type', None)
        
        # Map the block type to the card claimed
        if block_type is CoupActionType.BlockForeignAid:
            block_card = "Duke"
        elif block_type is CoupActionType.BlockStealCaptain:
            block_card = "Captain"
        elif block_type is CoupActionType.BlockStealAmbassador:
            block_card = "Ambassador"
        elif block_type is CoupActionType.BlockAssassinate:
            block_card = "Contessa"
        else:
            raise ValueError(f"Unexpected block type: {block_type}")
        
        # Check if blocker actually has the card they claimed
        is_honest = self.state.game_state["hidden_hand"][blocker_player_id].count(block_card) > 0
        
        if is_honest:
            # Blocker was honest - they shuffle their card back and draw new, challenger loses a card
            self.state.game_state["hidden_hand"][blocker_player_id].remove(block_card)
            self.state.game_state["pile"].append(block_card)
            random.shuffle(self.state.game_state["pile"])
            
            new_card = self.state.game_state["pile"].pop()
            self.state.game_state["hidden_hand"][blocker_player_id].append(new_card)
            
            # Challenger loses a card
            card_lost = self._make_player_lose_a_card(challenger_player_id)
            challenger_remaining = len(self.state.game_state["hidden_hand"][challenger_player_id])
            
            # Send observations
            blocker_msg = f"You were challenged on your {block_card} block by Player #{challenger_player_id}. Since you had the {block_card}, you shuffled it back and drew a {new_card}. Player #{challenger_player_id} lost a {card_lost} and has {challenger_remaining} card(s) remaining."
            challenger_msg = f"Your challenge on Player #{blocker_player_id}'s {block_card} block failed. They did have a {block_card}. You lost a {card_lost} and have {challenger_remaining} card(s) remaining."
            others_msg = f"Player #{challenger_player_id} challenged Player #{blocker_player_id}'s {block_card} block and failed. Player #{blocker_player_id} had the {block_card}, shuffled it back and drew a new card. Player #{challenger_player_id} lost a {card_lost} and has {challenger_remaining} card(s) remaining."
            
            # The block was successful, so the original action is cancelled
            self.state.game_state["action_metadata"].action_type = CoupActionType.PASS
            
        else:
            # Blocker was lying - they lose a card, original action proceeds
            card_lost = self._make_player_lose_a_card(blocker_player_id)
            blocker_remaining = len(self.state.game_state["hidden_hand"][blocker_player_id])
            
            # Send observations
            blocker_msg = f"You were challenged on your {block_card} block by Player #{challenger_player_id}. Since you didn't have a {block_card}, you lost a {card_lost} and have {blocker_remaining} card(s) remaining."
            challenger_msg = f"Your challenge on Player #{blocker_player_id}'s {block_card} block succeeded! They didn't have a {block_card}. They lost a {card_lost} and have {blocker_remaining} card(s) remaining."
            others_msg = f"Player #{challenger_player_id} successfully challenged Player #{blocker_player_id}'s {block_card} block. Player #{blocker_player_id} didn't have the {block_card}, lost a {card_lost} and has {blocker_remaining} card(s) remaining."
            
            # The block failed, so the original action will proceed
            # No need to change action_type
        
        # Send all observations
        self.state.add_observation(from_id=ta.GAME_ID, to_id=blocker_player_id, message=blocker_msg, for_logging=False)
        self.state.add_observation(from_id=ta.GAME_ID, to_id=challenger_player_id, message=challenger_msg, for_logging=False)
        self._broadcast_observations(others_msg, exclude_player_ids=[blocker_player_id, challenger_player_id])
        
        # Mark that we're done querying
        self.state.game_state["action_metadata"].players_to_query = None

    def _execute_exchange_action(self):
        """Execute the exchange action after player has chosen which cards to keep"""
        metadata = self.state.game_state["action_metadata"]
        player_id = metadata.source_player_id
        cards_to_keep = metadata.cards_to_keep
        
        # Get all cards currently in hand (includes the 2 drawn cards)
        all_cards = self.state.game_state["hidden_hand"][player_id][:]
        
        # Validate that the cards to keep are actually available
        cards_available = all_cards[:]
        for card in cards_to_keep:
            if card in cards_available:
                cards_available.remove(card)
            else:
                raise ValueError(f"Cannot keep {card} - it's not one of your available cards")
        
        # Update player's hand with only the kept cards
        self.state.game_state["hidden_hand"][player_id] = list(cards_to_keep)
        
        # Put the non-kept cards back in the pile and shuffle
        for card in cards_available:
            self.state.game_state["pile"].append(card)
        random.shuffle(self.state.game_state["pile"])
        
        # Send observations
        if len(cards_to_keep) == 1:
            source_observation = f"You have completed your exchange and kept: {cards_to_keep[0]}"
        else:
            source_observation = f"You have completed your exchange and kept: {', '.join(cards_to_keep)}"
        other_observation = f"Player #{player_id} has completed their exchange."
        
        self.state.add_observation(from_id=ta.GAME_ID, to_id=player_id, message=source_observation, for_logging=False)
        self._broadcast_observations(other_observation, exclude_player_ids=[player_id])

    def _broadcast_observations(self, other_player_observations: str, exclude_player_ids: Optional[List[int]] = None):
        """ Broadcast the observations to the players except for the players in `exclude_player_ids` """
        for pid in range(self.state.num_players):
            if pid in exclude_player_ids:
                continue
            self.state.add_observation(from_id=ta.GAME_ID, to_id=pid, message=other_player_observations, for_logging=False)

    # ---------------------------------------------------------------------
    # PROMPT GENERATION METHODS -- CONVERTS GAME STATE TO PROMPT
    # ---------------------------------------------------------------------
    def _gen_initial_prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        prompt = base_coup_prompts.base_prompt.replace("<NUM_PLAYERS>", str(self.state.num_players))
        prompt = prompt.replace("<PLAYER_ID>", str(player_id))
        prompt = prompt.replace("<PLAYER_OBSERVATIONS>", self._make_player_observations_prompt(player_id))
        return prompt
    

    def _make_last_action_msg(self, player_id: int, action: CoupActionType) -> str:
        """
        Make a prompt for the last action that was taken.
        Remember: This function will never be called when player_id is the current player
        """
        # TODO: re-check this thing with the new phases
        game_state = self.state.game_state
        curr_pid = self.state.current_player_id
        
        # If the game is in challenge phase and current_player==source, it means the current_player just made a challengable claim and we are now going to be asked to block or call bullshit
        if game_state["phase"] == "challenge" and curr_pid == game_state["last_action_source_player_id"]:
            action = game_state["last_action"]
            action_target_player_id = game_state["last_action_target_player_id"]
            you_str = " (you)" if action_target_player_id == player_id else ""
            affected_player_str = f" on Player {action_target_player_id}{you_str}." if action_target_player_id is not None and action not in {CoupActionType.Exchange, CoupActionType.Tax, CoupActionType.ForeignAid, CoupActionType.Keep} else "."
            claim_str = f" They're claiming they have a {self._action_to_card(action)} card." if action != CoupActionType.ForeignAid else ""
            return f"Player #{curr_pid} just played {action.value}{affected_player_str}{claim_str}"
        elif game_state["phase"] == "challenge" and curr_pid != game_state["last_action_source_player_id"]:
            action_str = f"call [BULLSHIT] on" if game_state["last_action"] == CoupActionType.BULLSHIT else f"[PASS] for their turn regarding challenging"
            your_str = " (your)" if game_state['last_action_source_player_id'] == player_id else ""
            return f"Player #{curr_pid} has decided to {action_str} Player #{game_state['last_action_source_player_id']}'s {your_str} claim."
        
        elif action is CoupActionType.Keep:   # Special exception for the exchange action
            return f"Player #{curr_pid} just completed their exchange."
        else: # game_state["phase"] == "play", so we are in play mode not challenge mode, and we know the action is not Keep
            return f"Player #{curr_pid} just played {action.value}."

    def _make_player_observations_prompt(self, player_id: Optional[int] = None) -> str:
        """
        Make a prompt for the specified player. Tells them their latest hand and the state of the game (what cards are in the pile, what cards are in the revealed hand, etc).
        """
        if player_id is None:
            player_id = self.state.current_player_id
        game_state = self.state.game_state
        msg = f"There are {self.state.num_players} players in the game.\n"

        for pid in range(self.state.num_players):
            if pid == player_id: 
                continue
            if len(game_state['hidden_hand'][pid]) == 0:
                msg += f"Player #{pid} is out.\n"
            elif len(game_state['revealed_hand'][pid]) > 0:
                msg += f"Player #{pid} has {game_state['coins'][pid]} coins, has revealed and lost a {game_state['revealed_hand'][pid][0]} card, and has {len(game_state['hidden_hand'][pid])} hidden influence cards remaining.\n"
            else:
                msg += f"Player #{pid} has {game_state['coins'][pid]} coins, and has {len(game_state['hidden_hand'][pid])} hidden influence cards remaining.\n"


        msg += f"\n ------ You are Player #{player_id}. You have {game_state['coins'][player_id]} coins"

        if len(game_state['hidden_hand'][player_id]) == 2:
            msg += f" and {len(game_state['hidden_hand'][player_id])} hidden influence cards remaining: You have a {game_state['hidden_hand'][player_id][0]} and a {game_state['hidden_hand'][player_id][1]}."
        elif len(game_state['hidden_hand'][player_id]) == 1:
            msg += f", a hidden {game_state['hidden_hand'][player_id][0]} card, and you have a revealed {game_state['revealed_hand'][player_id][0]} card that is out of play."

        msg += " --------\n"
        return msg
    
    def _send_call_to_action_prompt(self) -> None:
        """
        Make a prompt for the player that asks them to make an action or challenge. 
        This is only called when the player IS the current_player, done AFTER we've advanced the turn.
        """
        msg = base_coup_prompts.base_reprompt
        msg = msg.replace("<PLAYER_OBSERVATIONS>", self._make_player_observations_prompt())
        
        # Determine the call to action based on the current phase
        if self.state.game_state["phase"] == GamePhase.QueryWhichToKeep:
            # Player needs to choose which cards to keep after exchange
            call_to_action_str = f"You need to choose which {'two cards' if len(self.state.game_state['revealed_hand'][self.state.current_player_id]) == 0 else 'card'} to keep. Use [keep <card1>" + \
                f"{' <card2>' if len(self.state.game_state['revealed_hand'][self.state.current_player_id]) == 0 else ''}]"
            
        elif self.state.game_state["phase"] == GamePhase.QueryForBlockOrChallenge:
            # Player is being asked if they want to block or challenge
            metadata = self.state.game_state["action_metadata"]
            action = metadata.action_type
            source_player_id = metadata.source_player_id
            target_player_id = metadata.target_player_id
            
            # Build the action description
            action_desc = f"Player #{source_player_id} is attempting to {action.value}"
            if action is CoupActionType.Assassinate:
                action_desc += " (they have paid 3 coins)"
            if target_player_id is not None and target_player_id != self.state.current_player_id:
                action_desc += f" on Player #{target_player_id}"
            elif target_player_id == self.state.current_player_id:
                action_desc += " on you"
            
            # Determine valid block options for current player
            block_options = []
            if action is CoupActionType.ForeignAid:
                block_options.append("[block foreign aid]")
            elif action is CoupActionType.Steal and target_player_id == self.state.current_player_id:
                block_options.extend(["[block steal captain]", "[block steal ambassador]"])
            elif action is CoupActionType.Assassinate and target_player_id == self.state.current_player_id:
                block_options.append("[block assassinate]")
            
            # Build the call to action
            if block_options:
                block_str = ", ".join(block_options) + ", "
            else:
                block_str = ""
            
            # Add claim info for challengeable actions
            claim_str = ""
            if action in {CoupActionType.Tax, CoupActionType.Assassinate, CoupActionType.Steal, CoupActionType.Exchange}:
                claim_str = f" (claiming {self._action_to_card(action)})"
            
            # Build the call to action with proper handling for foreign aid
            bullshit_str = "call [BULLSHIT], " if action is not CoupActionType.ForeignAid else ""
            call_to_action_str = f"{action_desc}{claim_str}. Do you want to {block_str}{bullshit_str}or [PASS]?"
            
        elif self.state.game_state["phase"] == GamePhase.QueryToChallengeTheBlocker:
            # Someone blocked, asking if anyone wants to challenge the block
            metadata = self.state.game_state["action_metadata"]
            blocker_id = metadata.blocker_player_id
            block_type = metadata.block_type
            
            # Determine which card the blocker is claiming based on block type
            if block_type is CoupActionType.BlockForeignAid:
                block_claim = "Duke"
            elif block_type is CoupActionType.BlockStealCaptain:
                block_claim = "Captain"
            elif block_type is CoupActionType.BlockStealAmbassador:
                block_claim = "Ambassador"
            elif block_type is CoupActionType.BlockAssassinate:
                block_claim = "Contessa"
            else:
                block_claim = "unknown card"
            
            call_to_action_str = f"Player #{blocker_id} is blocking with {block_claim}. Do you want to call [BULLSHIT] or [PASS]?"
            
        elif self.state.game_state["coins"][self.state.current_player_id] >= 10:
            # Forced coup
            call_to_action_str = "You have 10 or more coins and must coup. Use [coup x] where x is the player id number."
            
        else:
            # Normal play phase
            call_to_action_str = "What action do you want to take?"
        
        msg = msg.replace("<CALL_TO_ACTION_OR_CHALLENGE>", call_to_action_str)
        
        # Send the message to the current player
        self.state.add_observation(from_id=ta.GAME_ID, to_id=self.state.current_player_id, message=msg, for_logging=False)

    def _action_to_card(self, action: CoupActionType) -> str:
        """ Convert a CoupActionType to a card """
        
        if action is CoupActionType.Tax:
            return "Duke"
        elif action is CoupActionType.Assassinate:
            return "Assassin"
        elif action is CoupActionType.Steal:
            return "Captain"
        elif action is CoupActionType.Exchange:
            return "Ambassador"
        
        elif action is CoupActionType.BlockAssassinate:
            return "Contessa"
        elif action is CoupActionType.BlockStealCaptain:
            return "Captain"
        elif action is CoupActionType.BlockStealAmbassador:
            return "Ambassador"
        elif action is CoupActionType.BlockForeignAid:
            return "Duke"
        
        else:
            return "."

    # ---------------------------------------------------------------------
    # GAME STATE ADJUSTMENT METHODS -- SYNTACTIC SUGAR FOR THE GAME LOGIC
    # ---------------------------------------------------------------------

    def _make_player_lose_a_card(self, player_id: int):
        """
        In a perfect world, the player would pick which card they want to lose.
        But for now, we'll just pop the last card in their hand.
        """
        card = self.state.game_state["hidden_hand"][player_id].pop()
        self.state.game_state["revealed_hand"][player_id].append(card)
        return card

    def _advance_turn(self):
        """
        Advance the state to the next player based on current phase.
        """
        if self.state.game_state["phase"] == GamePhase.Play:
            # Normal play phase - go to next player in turn order
            current_pid = self.state.current_player_id
            next_pid = (current_pid + 1) % self.state.num_players
            # Skip eliminated players
            while len(self.state.game_state["hidden_hand"][next_pid]) == 0:
                next_pid = (next_pid + 1) % self.state.num_players
                if next_pid == current_pid:  # All other players eliminated
                    break

            # Income and Coup immediately advance the turn, so action_metadata is no longer needed
            if self.state.game_state["action_metadata"] is not None and self.state.game_state["action_metadata"].action_type in {CoupActionType.Income, CoupActionType.Coup}:
                self.state.game_state["action_metadata"] = None
                    
        elif self.state.game_state["phase"] == GamePhase.QueryWhichToKeep:
            # Check if the exchange has been completed (cards_to_keep is set)
            metadata = self.state.game_state["action_metadata"]
            if hasattr(metadata, 'cards_to_keep') and metadata.cards_to_keep is not None:
                # Exchange is complete, move to next player
                self.state.game_state["phase"] = GamePhase.Play
                source_pid = metadata.source_player_id
                next_pid = (source_pid + 1) % self.state.num_players
                # Skip eliminated players
                while len(self.state.game_state["hidden_hand"][next_pid]) == 0:
                    next_pid = (next_pid + 1) % self.state.num_players
                    if next_pid == source_pid:  # All other players eliminated
                        break
                self.state.game_state["action_metadata"] = None
            else:
                # Stay with the same player - they still need to choose cards
                next_pid = self.state.current_player_id
            
        elif self.state.game_state["phase"] in {GamePhase.QueryForBlockOrChallenge, GamePhase.QueryToChallengeTheBlocker}:
            metadata = self.state.game_state["action_metadata"]
            if metadata.players_to_query and len(metadata.players_to_query) > 0:
                # Query next player in the list
                next_pid = metadata.players_to_query.pop(0)
            else:
                # No more players to query
                if self.state.game_state["phase"] == GamePhase.QueryToChallengeTheBlocker:
                    # Only cancel the original action if no one challenged the block
                    # (If someone did challenge and the blocker was honest, the action was already set to PASS in _execute_showdown_on_blocker_bullshit)
                    if not hasattr(metadata, 'blocker_challenger_player_id') or metadata.blocker_challenger_player_id is None:
                        # Block was not challenged, so cancel the original action
                        self.state.game_state["action_metadata"].action_type = CoupActionType.PASS
                
                # Execute the action (may be PASS if blocked/challenged successfully)
                self._execute_current_action()
                
                # If we just executed an Exchange, we'll be in QueryWhichToKeep phase
                if self.state.game_state["phase"] == GamePhase.QueryWhichToKeep:
                    next_pid = self.state.game_state["action_metadata"].source_player_id
                    self.state.game_state["action_metadata"].players_to_query = [] # No more players to query
                else:
                    # Return to normal play - next player after action source
                    self.state.game_state["phase"] = GamePhase.Play
                    source_pid = metadata.source_player_id
                    next_pid = (source_pid + 1) % self.state.num_players
                    # Skip eliminated players
                    while len(self.state.game_state["hidden_hand"][next_pid]) == 0:
                        next_pid = (next_pid + 1) % self.state.num_players
                        if next_pid == source_pid:  # All other players eliminated
                            break
                    self.state.game_state["action_metadata"] = None
        else:
            raise Exception(f"Unexpected game phase: {self.state.game_state['phase']}")
        
        self.state.manually_update_current_player(new_player_id=next_pid)
        
        # Check for winner
        winner = self._get_winner()
        if winner is not None:
            self.state.set_winners(player_ids=[winner], reason=f"Player {winner} has won the game!")
            return True, {"winner": winner}
        else:
            return False, {}

    def _get_winner(self) -> Optional[int]:
        """
        Check if the game is over. Return winning player id if so, otherwise return None.
        """
        remaining_players = [pid for pid in range(self.state.num_players) if len(self.state.game_state["hidden_hand"][pid]) > 0]
        if len(remaining_players) == 1:
            return remaining_players[0]
        else:
            return None


    # -----------------------------------------------------------------------
    #  PARSE ACTIONS -- JUST BOILERPLATE CODE NOTHING INTERESTING BELOW HERE
    # -----------------------------------------------------------------------
    def _parse_action(self, response_str: str) -> Tuple[CoupActionType, Optional[Union[int, List[str]]]]:
        """
        Extract the **last** square-bracket command from `response_str` and convert it to (CoupActionType, arg).

        The purpose of this method is TO PARSE ONLY. It makes sure the response is translated into a valid action, but does not validate against the state of the game.
        """
        # 1) Grab all bracketed chunks, take the last one.
        match_list = re.findall(r"\[([^\[\]]+)\]", response_str.replace("[GAME]", " "))
        if not match_list:
            raise ValueError(f"No bracketed command found. What is your desired [action]?")
        cmd = match_list[-1].strip().lower()
        tokens = cmd.split()

        if not tokens:
            raise ValueError("Empty command.")

        # ---------- Simple one-word actions ----------
        simple = {"income": CoupActionType.Income, "tax": CoupActionType.Tax,"exchange": CoupActionType.Exchange, "pass": CoupActionType.PASS, "bullshit": CoupActionType.BULLSHIT}
        if tokens[0] in simple:
            if len(tokens) > 1:
                raise ValueError(f"Invalid action: {tokens[0]}, cannot have more than one word when doing a {tokens[0]}")
            return simple[tokens[0]], None
        
        if tokens[:2] == ["foreign", "aid"]:
            if len(tokens) > 2:
                raise ValueError(f"Invalid action: {response_str}, cannot have more than two words when doing a foreign aid")
            return CoupActionType.ForeignAid, None

        # ---------- Directed actions ----------
        directed_map = {"coup": CoupActionType.Coup, "assassinate": CoupActionType.Assassinate, "steal": CoupActionType.Steal }
        if tokens[0] in directed_map:
            if len(tokens) < 2 or not tokens[1].isdigit():
                raise ValueError(f"Missing / invalid target for '{tokens[0]}'.")
            return directed_map[tokens[0]], int(tokens[1])

        # ---------- Ambassador "keep" special ----------
        if tokens[0] == "keep":
            if len(tokens) < 2 or len(tokens) > 3:
                raise ValueError("'keep' must specify one or two cards to keep.")
            
            cards = []
            for i in range(1, len(tokens)):
                if tokens[i].lower() not in {"duke", "assassin", "ambassador", "captain", "contessa"}:
                    raise ValueError(f"Invalid card name: {tokens[i]}")
                cards.append(tokens[i].lower())
            
            return CoupActionType.Keep, cards

        # ---------- Blocks ----------
        if tokens[0] == "block":
            if tokens[1:3] == ["foreign", "aid"]:
                return CoupActionType.BlockForeignAid, None
            if tokens[1] == "steal" and len(tokens) >= 3:
                if tokens[2] == "ambassador":
                    return CoupActionType.BlockStealAmbassador, None
                if tokens[2] == "captain":
                    return CoupActionType.BlockStealCaptain, None
                raise ValueError("Block steal must specify 'ambassador' or 'captain'.")
            if tokens[1] == "assassinate":
                return CoupActionType.BlockAssassinate, None

        raise ValueError(f"Unrecognized command: [{cmd}]")
    

    #########################################################################
    #  RENDERING METHOD
    #########################################################################
    def _game_state_to_headline_str(self, game_state: Optional[Dict[str, Any]] = None):
        """
        Convert the game state to a headline string.
        """
        if hasattr(self.state, "game_state") and self.state.game_state is not None and self._get_winner() == self.state.current_player_id:
            return f"Player #{self._get_winner()} has won!"
        if game_state["phase"] == GamePhase.Play:
            return f"It's Player #{self.state.current_player_id}'s turn"
        elif game_state["phase"] == GamePhase.QueryForBlockOrChallenge:
            tgt_player_str = ""
            if game_state['action_metadata'].action_type in {CoupActionType.Assassinate, CoupActionType.Steal}:
                tgt_player_str = f" on Player #{game_state['action_metadata'].target_player_id}" if game_state['action_metadata'].action_type is CoupActionType.Assassinate else f" from Player #{game_state['action_metadata'].target_player_id}"
            return f"Player #{game_state['action_metadata'].source_player_id} is attempting a {game_state['action_metadata'].action_type.name}{tgt_player_str}, asking if Player #{self.state.current_player_id} wants to block/challenge."
        elif game_state["phase"] == GamePhase.QueryToChallengeTheBlocker:
            return f"Player #{game_state['action_metadata'].blocker_player_id} is doing a {game_state['action_metadata'].block_type.name} on Player #{game_state['action_metadata'].source_player_id}, asking if Player #{self.state.current_player_id} wants to challenge the block."
        elif game_state["phase"] == GamePhase.QueryWhichToKeep:
            return f"Player #{game_state['action_metadata'].source_player_id} is attempting an Exchange, asking which they wish to keep."
        else:
            raise Exception(f"Unexpected game phase: {game_state['phase']}")
        
    def _get_player_marker(self, player_id: int, game_state: Optional[Dict[str, Any]] = None):
        """
        Helper function to display a useful marker to show the player's status in the game.
        """
        if hasattr(self.state, "game_state") and self.state.game_state is not None and self._get_winner() == player_id:
            return "👑"  # Player is the winner
        if game_state["hidden_hand"][player_id] == []:
            return "💀"  # Player is eliminated
        if (game_state["phase"] == GamePhase.QueryForBlockOrChallenge or game_state["phase"] == GamePhase.QueryToChallengeTheBlocker) and \
            player_id == game_state["action_metadata"].blocker_player_id:
            return "B"  # Player is blocking the action
        
        if game_state["phase"] == GamePhase.QueryWhichToKeep and player_id == game_state["action_metadata"].source_player_id:
            return "🔀"  # Player being asked which cards they wish to keep
        
        if self.state.current_player_id == player_id:
            return "*"  # Player is the current player
        
        if game_state["action_metadata"] is not None and game_state["action_metadata"].source_player_id == player_id:
            return "."  # Player is awaiting potential challenges
        
        return " "  # Player is not involved in the action

    def _render_board(self, game_state: Optional[Dict[str, Any]] = None):
        """
        Pretty console renderer for a Coup game_state.

        The offset and adjustment are used to ensure that the board is always LINE_WIDTH characters wide, because ansii codes add to char count but not to spacing.
        """
        # --- helpers -----------------------------------------------------------
        game_state = self.state.game_state if game_state is None else game_state
        
        # Card colors - specific ANSI codes for each card type
        CARD_COLOURS = {
            "Contessa": 91,      # Red
            "Ambassador": 92,    # Green
            "Duke": 95,          # Purple/Magenta
            "Captain": 94,       # Blue
            "Assassin": 90       # Dark Grey
        }
        
        # Player header colors - using different colors from cards
        # Using: Yellow(93), Cyan(96), White(97), and variations
        PLAYER_COLOURS = [93, 96, 97, 33, 35, 36]

        def colour(text: str, code: int) -> str:
            return f"\033[{code}m{text}\033[0m"

        def bold_underline_colour(text: str, code: int) -> str:
            # Combines bold (1), underline (4) and color in one ANSI code
            return f"\033[1;4;{code}m{text}\033[0m"

        def strike(text: str) -> str:
            # ANSI strike-through (not supported in a few older terminals)
            return f"\033[9m{text}\033[0m"


        # -----------------------------------------------------------------------

        out_lines = ["",self._game_state_to_headline_str(game_state), ""]
        # Sort players numerically for reproducibility
        for pid in range(self.state.num_players):
            player_colour_code = PLAYER_COLOURS[pid % len(PLAYER_COLOURS)]

            # "*" means this is the current player who just did an action
            # "." means this is the player who initially triggered a QueryForX phase
            # " " means this is a player who is not involved in the action

            # ----- header: "Player #x" (unique colour) -----
            curr_player_marker = self._get_player_marker(pid, game_state)
            header = bold_underline_colour(f"[{curr_player_marker}] - Player #{pid}", player_colour_code)
            out_lines.append(f"{header}  ––  {game_state['coins'][pid]} coin{'s' if game_state['coins'][pid]!=1 else ''}")

            # ----- cards -----------------------------------
            hand_cards = game_state["hidden_hand"][pid]   # Hidden cards
            revealed   = game_state["revealed_hand"][pid] # Revealed/lost cards

            for card in hand_cards:
                card_colour_code = CARD_COLOURS.get(card, 97)  # Default to white if card not found
                out_lines.append(f"   - {colour(card, card_colour_code)}")

            # Sometimes revealed cards may no longer be in hand (fully lost)
            for card in revealed:
                card_colour_code = CARD_COLOURS.get(card, 97)  # Default to white if card not found
                out_lines.append(f"   - {colour(strike(card), card_colour_code)}")

            out_lines.append("")

        out_lines.append("")
        out_lines.append("")
        # from_ansi is used to ensure that the ANSI codes are properly padded while rendering
        return Text.from_ansi("\n".join(out_lines))

   


from typing import Dict, Any, List

def create_board_str(game_state: Dict[str, Any]) -> str:
    """
    Render the Lights Out board with lit cells shaded
    """
    board: List[List[int]] = game_state.get("board", [])
    if not board:
        return "Board not available."

    grid_size = len(board)
    on_symbol = "███"  
    off_symbol = "   "  

    def cell_display(r: int, c: int) -> str:
        """Returns the 3-character representation for a cell"""
        return on_symbol if board[r][c] == 1 else off_symbol

    def horizontal_line(left: str, mid: str, right: str) -> str:
        """Creates a horizontal line for the grid border"""
        return "  " + left + ("─" * 3 + mid) * (grid_size - 1) + "─" * 3 + right

    header = "    " + "   ".join(f"{c:^1}" for c in range(grid_size))
    lines = [header]
    lines.append(horizontal_line("┌", "┬", "┐"))

    for r in range(grid_size):
        row_content = "│".join(cell_display(r, c) for c in range(grid_size))
        lines.append(f"{r} │{row_content}│")
        if r < grid_size - 1:
            lines.append(horizontal_line("├", "┼", "┤"))

    lines.append(horizontal_line("└", "┴", "┘"))

    return "\n".join(lines)


import re, shutil, time, rich, rich.layout
from typing import Dict, Optional, Tuple
from textarena.core import Env, Message, Info, RenderWrapper, State

__all__ = ["SimpleRenderWrapper"]

class SimpleRenderWrapper(RenderWrapper):
    def __init__(self, env: Env, player_names: Optional[Dict[int, str]] = None, render_mode: str = "multi"):
        super().__init__(env)
        self.player_names = player_names
        self.render_mode = render_mode
        assert render_mode in ["standard", "board", "chat", "multi"], \
            f"The selected render_mode does not exist. The available options are:" +\
            f"\n\t'standard' - view both the game board and model chats"+\
            f"\n\t'board' - view just the game board"+\
            f"\n\t'chat' - view just the model chats side-by-side"+\
            f"\n\t'multi' - view the game board and a combined chat window"
        self.console = rich.console.Console()

    def _render(self):
        board = self.env.get_board_str() if hasattr(self.env, "get_board_str") and callable(getattr(self.env, "get_board_str")) else None
        logs = getattr(self.env.state, "logs", [])
        board = f"No game board provided by {self.env.env_id}\n(not implemented / not available)" if board is None else board
        board_panel = rich.panel.Panel.fit(board, title="Game Board", border_style="white", box=rich.box.SQUARE)

        # Separate logs by player
        logs_by_player = {}
        for pid, msg in logs:
            logs_by_player.setdefault(pid, []).append(msg)

        def get_message_text(pid, mode="standard", include_name=False):
            name = self.player_names.get(pid, f"Player {pid}")
            message_list = logs_by_player.get(pid, [])
            message = "(no message yet)" if not len(message_list) else message_list[-1].strip()

            terminal_size = shutil.get_terminal_size()
            if mode=="standard":
                max_chars = (terminal_size.columns/2-2)*(terminal_size.lines/3-3)
            elif mode=="chat":
                max_chars = (terminal_size.columns/2-2)*(terminal_size.lines-3)
            elif mode=="multi":
                pid = [i for i, msg in logs if i!=-1][-1] if len(logs) else None
                name = self.player_names.get(pid, f"Player {pid}")
                message = "(no message yet)" if pid is None else logs_by_player.get(pid)[-1].strip()
                max_chars = (terminal_size.columns-2)*(terminal_size.lines/3-3)
            
            # truncate message
            message = message[-int(max_chars-len(name)-3):]
            return rich.panel.Panel(rich.text.Text(f"[{name}] {message}" if include_name else message, no_wrap=False, overflow="fold"), title=name, border_style="white")

        # Setup layout
        layout = rich.layout.Layout()

        if self.render_mode == "standard":
            layout.split_column(rich.layout.Layout(name="spacer", size=1), rich.layout.Layout(name="top", ratio=2), rich.layout.Layout(name="bottom", ratio=1))
            layout["top"].update(rich.align.Align.center(board_panel, vertical="middle"))
            layout["bottom"].split_row(rich.layout.Layout(name="chat0"), rich.layout.Layout(name="chat1"))
            layout["chat0"].update(get_message_text(0, "standard"))
            layout["chat1"].update(get_message_text(1, "standard"))

        elif self.render_mode == "board":
            layout.update(rich.align.Align.center(board_panel, vertical="middle"))

        elif self.render_mode == "chat":
            layout.split_column(rich.layout.Layout(name="spacer", size=1), rich.layout.Layout(name="chats", ratio=1))
            layout["chats"].split_row(rich.layout.Layout(name="chat0"), rich.layout.Layout(name="chat1"))
            layout["chat0"].update(get_message_text(0, "chat"))
            layout["chat1"].update(get_message_text(1, "chat"))

        elif self.render_mode == "multi":
            layout.split_column(rich.layout.Layout(name="spacer", size=1), rich.layout.Layout(name="top", ratio=2), rich.layout.Layout(name="bottom", ratio=1))
            layout["top"].update(rich.align.Align.center(board_panel, vertical="middle"))
            layout["bottom"].update(get_message_text(None, "multi"))

        self.console.clear()
        self.console.print(layout)

    def reset(self, num_players: int, seed: Optional[int]=None) -> None:
        result = self.env.reset(num_players=num_players, seed=seed)
        self.state = self.env.state
        if self.player_names is None:
            self.player_names = {pid: f"Player {pid}" for pid in range(self.state.num_players)}
        self.player_names.update(self.state.role_mapping)
        self.game_over = False

        # assert render mode with num players
        if self.render_mode in ["standard", "chat"]:
            assert num_players==2, f"render_modes 'standard' and 'chat' can only be used with two players"
        return result

    def step(self, action: str) -> Tuple[bool, Optional[Info]]:
        step_results = self.env.step(action=action)
        time.sleep(0.2)
        self._render()
        time.sleep(0.2)
        return step_results

    def close(self):
        return self.env.close()


import random
import re
from typing import Optional, Tuple, List, Dict, Any

import textarena as ta

class BriscolaEnv(ta.Env):
    def __init__(self):
        """ Initializes the Briscola card game environment """
        super().__init__()
        self.deck = self._create_deck()
        
    def _create_deck(self) -> List[Dict[str, Any]]:
        """ Creates a 40-card Italian deck for Briscola """
        suits = ['♠', '♥', '♦', '♣']  # Spades, Hearts, Diamonds, Clubs
        ranks = ['A', '2', '3', '4', '5', '6', '7', 'J', 'Q', 'K']  # No 8, 9, 10 in Italian deck
        deck = []
        
        for suit in suits:
            for rank in ranks:
                card = {
                    'rank': rank,
                    'suit': suit,
                    'points': self._get_card_points(rank),
                    'power': self._get_card_power(rank)
                }
                deck.append(card)
        return deck
    
    def _get_card_points(self, rank: str) -> int:
        """ Returns the point value of a card in Briscola """
        point_values = {
            'A': 11,    # Ace
            '3': 10,    # Three
            'K': 4,     # King
            'Q': 3,     # Queen (Cavallo/Horse)
            'J': 2,     # Jack (Fante)
            '7': 0, '6': 0, '5': 0, '4': 0, '2': 0  # No points
        }
        return point_values[rank]
    
    def _get_card_power(self, rank: str) -> int:
        """ Returns the power/strength of a card for trick-taking (higher = stronger) """
        power_values = {
            'A': 8,     # Ace (strongest)
            '3': 7,     # Three
            'K': 6,     # King
            'Q': 5,     # Queen
            'J': 4,     # Jack
            '7': 3, '6': 2, '5': 1, '4': 0, '2': -1  # Weakest
        }
        return power_values[rank]
    
    def _card_to_string(self, card: Dict[str, Any]) -> str:
        """ Converts a card to a readable string """
        return f"{card['rank']}{card['suit']}"
    
    def _find_action_token(self, message: str) -> Optional[int]:
        """ Parse card play action from player message """
        pattern = re.compile(r"\[play (\d+)\]", re.I)
        match = pattern.search(message)
        
        if match:
            return int(match.group(1)) - 1  # Convert to 0-based index
        return None
    
    def reset(self, num_players: int = 2, seed: Optional[int] = None):
        """ Reset the game state """
        if num_players not in [2, 3, 4]:
            raise ValueError("Briscola supports 2, 3, or 4 players")
            
        if num_players == 2:
            self.state = ta.TwoPlayerState(num_players=num_players, seed=seed)
        else:
            self.state = ta.FFAMultiPlayerState(num_players=num_players, seed=seed)
        
        # Initialize game state
        game_state = {
            'players': {},
            'deck': [],
            'trump_suit': None,
            'trump_card': None,
            'current_trick': [],
            'tricks_won': {},
            'points_won': {},
            'phase': 'playing',  # playing, finished
            'trick_leader': 0,
            'cards_in_hand': 3 if num_players <= 3 else 2
        }
        
        self.state.reset(game_state=game_state, player_prompt_function=self._generate_player_prompt)
        self._init_game()
    
    def _init_game(self):
        """ Initialize the game """
        gs = self.state.game_state
        
        # Shuffle deck
        deck_copy = self.deck.copy()
        random.shuffle(deck_copy)
        
        # Deal cards to players
        cards_per_hand = gs['cards_in_hand']
        for player_id in range(self.state.num_players):
            player_hand = []
            for _ in range(cards_per_hand):
                if deck_copy:
                    player_hand.append(deck_copy.pop())
            
            gs['players'][player_id] = {
                'hand': player_hand,
                'points': 0
            }
            gs['tricks_won'][player_id] = []
            gs['points_won'][player_id] = 0
        
        # Set trump card (last card dealt becomes trump indicator)
        if deck_copy:
            gs['trump_card'] = deck_copy.pop()
            gs['trump_suit'] = gs['trump_card']['suit']
            deck_copy.insert(0, gs['trump_card'])  # Put trump card at bottom of deck
        
        gs['deck'] = deck_copy
        gs['current_trick'] = []
        gs['trick_leader'] = 0
        gs['phase'] = 'playing'
        
        # Announce game start
        self._announce_game_start()
        self._announce_turn(self.state.current_player_id)
    
    def _announce_game_start(self):
        """ Announce the start of the game with trump suit """
        gs = self.state.game_state
        trump_str = self._card_to_string(gs['trump_card']) if gs['trump_card'] else "None"
        
        self.state.add_observation(
            message=f"Briscola game started! Trump suit: {gs['trump_suit']} (Trump card: {trump_str})",
            observation_type=ta.ObservationType.GAME_MESSAGE
        )
    
    def _generate_player_prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        return (
            f"You are playing Briscola - Player {player_id}.\n"
            f"Goal: Win tricks and collect the most points (120 total points in the deck).\n"
            f"Card Points: A=11, 3=10, K=4, Q=3, J=2, others=0\n"
            f"Card Power: A > 3 > K > Q > J > 7 > 6 > 5 > 4 > 2\n"
            f"Trump cards beat non-trump cards regardless of power.\n\n"
            f"Action: '[play X]' where X is the position (1-{len(game_state['players'][player_id]['hand']) if player_id in game_state['players'] else 3}) of the card in your hand\n"
        )
    
    def _render_player_hand(self, player_id: int) -> str:
        """ Renders the player's hand """
        gs = self.state.game_state
        if player_id not in gs['players']:
            return "No cards"
            
        player = gs['players'][player_id]
        hand = player['hand']
        
        if not hand:
            return "No cards in hand"
        
        output = []
        output.append("Your hand:")
        for i, card in enumerate(hand):
            trump_indicator = " (TRUMP)" if card['suit'] == gs['trump_suit'] else ""
            output.append(f"  {i+1}. {self._card_to_string(card)} [{card['points']} pts]{trump_indicator}")
        
        return "\n".join(output)
    
    def _render_current_trick(self) -> str:
        """ Renders the current trick being played """
        gs = self.state.game_state
        if not gs['current_trick']:
            return "No cards played yet this trick."
        
        output = []
        output.append("Current trick:")
        for player_id, card in gs['current_trick']:
            trump_indicator = " (TRUMP)" if card['suit'] == gs['trump_suit'] else ""
            output.append(f"  Player {player_id}: {self._card_to_string(card)}{trump_indicator}")
        
        return "\n".join(output)
    
    def step(self, action: str) -> Tuple[bool, ta.Info]:
        player_id = self.state.current_player_id
        gs = self.state.game_state
        
        self.state.add_observation(
            from_id=player_id, 
            message=action, 
            observation_type=ta.ObservationType.PLAYER_ACTION
        )
        
        # Parse action
        card_index = self._find_action_token(action)
        
        if card_index is None:
            self.state.set_invalid_move("Use [play X] where X is the card position (1, 2, 3, etc.)")
            return self.state.step()
        
        # Validate card index
        player = gs['players'][player_id]
        if card_index < 0 or card_index >= len(player['hand']):
            self.state.set_invalid_move(f"Invalid card position. You have {len(player['hand'])} cards (1-{len(player['hand'])})")
            return self.state.step()
        
        # Play the card
        played_card = player['hand'].pop(card_index)
        gs['current_trick'].append((player_id, played_card))
        
        self.state.add_observation(
            message=f"Player {player_id} played {self._card_to_string(played_card)}",
            observation_type=ta.ObservationType.GAME_ACTION_DESCRIPTION
        )
        
        # Check if trick is complete
        if len(gs['current_trick']) == self.state.num_players:
            return self._resolve_trick()
        else:
            # Move to next player
            return self._next_player_in_trick()
    
    def _next_player_in_trick(self) -> Tuple[bool, ta.Info]:
        """ Move to the next player in the current trick """
        next_player = (self.state.current_player_id + 1) % self.state.num_players
        self.state.manually_set_current_player_id(next_player)
        
        self._announce_turn(next_player)
        return self.state.step(rotate_player=False)
    
    def _resolve_trick(self) -> Tuple[bool, ta.Info]:
        """ Resolve the completed trick """
        gs = self.state.game_state
        
        # Determine trick winner
        winner_id, winning_card = self._determine_trick_winner(gs['current_trick'], gs['trump_suit'])
        
        # Calculate points in this trick
        trick_points = sum(card['points'] for _, card in gs['current_trick'])
        
        # Award points and trick to winner
        gs['points_won'][winner_id] += trick_points
        gs['tricks_won'][winner_id].append(gs['current_trick'].copy())
        
        self.state.add_observation(
            message=f"Player {winner_id} wins the trick with {self._card_to_string(winning_card)} and gains {trick_points} points!",
            observation_type=ta.ObservationType.GAME_MESSAGE
        )
        
        # Clear current trick
        gs['current_trick'] = []
        
        # Deal new cards if deck has cards
        self._deal_new_cards()
        
        # Check for game end
        if self._is_game_over():
            return self._end_game()
        
        # Winner leads next trick
        gs['trick_leader'] = winner_id
        self.state.manually_set_current_player_id(winner_id)
        
        self._announce_turn(winner_id)
        return self.state.step(rotate_player=False)
    
    def _determine_trick_winner(self, trick: List[Tuple[int, Dict]], trump_suit: str) -> Tuple[int, Dict]:
        """ Determine who wins the trick """
        if not trick:
            return 0, {}
        
        # Get the lead card (first card played) and lead suit
        lead_player_id, lead_card = trick[0]
        lead_suit = lead_card['suit']
        
        # Separate trump cards from non-trump cards
        trump_cards = [(pid, card) for pid, card in trick if card['suit'] == trump_suit]
        
        # Rule 1: If there are trump cards, highest trump wins
        if trump_cards:
            winner_id, winning_card = max(trump_cards, key=lambda x: x[1]['power'])
            return winner_id, winning_card
        
        # Rule 2: No trump cards played
        # Find all cards that follow the lead suit
        lead_suit_cards = [(pid, card) for pid, card in trick if card['suit'] == lead_suit]
        
        if lead_suit_cards:
            # If there are cards following suit, highest power of lead suit wins
            winner_id, winning_card = max(lead_suit_cards, key=lambda x: x[1]['power'])
            return winner_id, winning_card
        else:
            # This shouldn't happen in normal Briscola play since the lead card 
            # should always be in lead_suit_cards, but as a safety fallback:
            return lead_player_id, lead_card
    
    def _deal_new_cards(self):
        """ Deal new cards to players after a trick """
        gs = self.state.game_state
        
        if not gs['deck']:
            return
        
        # Deal one card to each player, starting with trick winner
        players_to_deal = []
        start_player = gs['trick_leader']
        
        for i in range(self.state.num_players):
            player_id = (start_player + i) % self.state.num_players
            players_to_deal.append(player_id)
        
        for player_id in players_to_deal:
            if gs['deck'] and len(gs['players'][player_id]['hand']) < gs['cards_in_hand']:
                new_card = gs['deck'].pop()
                gs['players'][player_id]['hand'].append(new_card)
    
    def _is_game_over(self) -> bool:
        """ Check if the game is over """
        gs = self.state.game_state
        
        # Game is over when all players have no cards left
        for player in gs['players'].values():
            if player['hand']:
                return False
        
        return True
    
    def _end_game(self) -> Tuple[bool, ta.Info]:
        """ End the game and determine winner """
        gs = self.state.game_state
        gs['phase'] = 'finished'
        
        # Find winner (most points)
        winner_id = max(gs['points_won'].keys(), key=lambda pid: gs['points_won'][pid])
        winner_points = gs['points_won'][winner_id]
        
        # Create final summary
        summary = f"Game Over! Player {winner_id} wins with {winner_points} points!\n\nFinal Scores:\n"
        sorted_players = sorted(gs['points_won'].items(), key=lambda x: x[1], reverse=True)
        
        for player_id, points in sorted_players:
            tricks_count = len(gs['tricks_won'][player_id])
            summary += f"Player {player_id}: {points} points ({tricks_count} tricks)\n"
        
        if self.state.num_players == 2:
            # Set winner for two-player game
            self.state.set_winner(winner_id, summary)
        else:
            # For multiplayer, set outcome
            self.state.set_outcome(
                reward=1 if self.state.current_player_id == winner_id else 0,
                reason=summary
            )
        
        return self.state.step(rotate_player=False)
    
    def _announce_turn(self, player_id: int):
        """ Announce the current player's turn """
        gs = self.state.game_state
        
        hand_str = self._render_player_hand(player_id)
        trick_str = self._render_current_trick()
        
        # Show current scores
        scores = []
        for pid in range(self.state.num_players):
            scores.append(f"Player {pid}: {gs['points_won'][pid]} pts")
        scores_str = " | ".join(scores)
        
        trump_info = f"Trump suit: {gs['trump_suit']}"
        if gs['deck']:
            trump_info += f" | Cards left in deck: {len(gs['deck'])}"
        
        message = f"{hand_str}\n\n{trick_str}\n\nScores: {scores_str}\n{trump_info}\n\nPlay a card using [play X]"
        
        self.state.add_observation(
            to_id=player_id,
            message=message,
            observation_type=ta.ObservationType.GAME_BOARD
        )
    
    def get_board_str(self) -> str:
        """ Get a string representation of the current game state """
        gs = self.state.game_state
        if not gs:
            return "Game not started"
            
        output = []
        output.append("=== BRISCOLA GAME ===")
        output.append(f"Trump suit: {gs['trump_suit']}")
        
        if gs['trump_card']:
            output.append(f"Trump card: {self._card_to_string(gs['trump_card'])}")
        
        output.append(f"Cards left in deck: {len(gs['deck'])}")
        output.append("")
        
        # Current trick
        if gs['current_trick']:
            output.append("Current trick:")
            for player_id, card in gs['current_trick']:
                trump_indicator = " (TRUMP)" if card['suit'] == gs['trump_suit'] else ""
                output.append(f"  Player {player_id}: {self._card_to_string(card)}{trump_indicator}")
            output.append("")
        
        # Player information
        for player_id in range(self.state.num_players):
            if player_id in gs['players']:
                player = gs['players'][player_id]
                hand_size = len(player['hand'])
                points = gs['points_won'][player_id]
                tricks = len(gs['tricks_won'][player_id])
                
                output.append(f"Player {player_id}: {points} points, {tricks} tricks, {hand_size} cards in hand")
        
        return "\n".join(output)


import textarena as ta 
from textarena.core import ObservationWrapper, Env, Observations, Info
from typing import Dict, Optional, Tuple, List

__all__ = ["ClassicalReasoningEvalsObservationWrapper"]


class ClassicalReasoningEvalsObservationWrapper(ObservationWrapper):
    """ TODO """
    def __init__(self, env: Env):
        super().__init__(env)
        self.full_observations: Dict[int, List[Tuple[int, str]]] = {}


    def _convert_obs_to_str(self, player_id: int) -> Observations:
        """ TODO """
        return (
            f"<｜begin▁of▁sentence｜>Please reason step by step, and put your final answer within \\boxed{{}}."
            f"<｜User｜>{self.full_observations[player_id][-1][1]}"
            "<｜Assistant｜><think>\n"
        )


    def observation(self, player_id: int, observation: Optional[ta.Observations]):
        if observation is None:
            return ""

            
        # Extend the full observations with the current observations without duplicates
        if player_id not in self.full_observations:
            self.full_observations[player_id] = []

        # Append new observations in sequence
        self.full_observations[player_id].extend(observation)

        return self._convert_obs_to_str(player_id=player_id) 



from typing import Dict, Any, Tuple

def create_board_str(game_state: Dict[str, Any]) -> str:
    board = game_state.get("board", [])
    scores = game_state.get("scores", {})
    grid_size = len(board)
    
    # Collect matched positions
    matched_positions = set()
    for r in range(grid_size):
        for c in range(grid_size):
            if board[r][c] != ".":
                matched_positions.add((r, c))

    # Header row
    output = ["🧠 Memory Game", "  " + " ".join(f"{c:2}" for c in range(grid_size))]

    # Board rows
    for r in range(grid_size):
        row = f"{r:2} "
        for c in range(grid_size):
            cell = board[r][c]
            if (r, c) in matched_positions:
                row += f"{cell:2} "
            else:
                row += "🔲 "
        output.append(row)

    # Scores
    output.append("\n📊 Scores:")
    for player_id, score_obj in scores.items():
        output.append(f"Player {player_id}: {score_obj['Score']} point(s)")

    return "\n".join(output)



import re
from typing import Any, Dict, Optional, Tuple, List

import textarena as ta
from textarena.envs.Checkers.renderer import create_board_str

class CheckersEnv(ta.Env):
    def __init__(self, max_turns: int = 50):
        """
        Args:
            max_turns (int): Maximum number of turns before the game ends in a draw.
        """
        self.max_turns = max_turns

    def get_board_str(self): return create_board_str(game_state=self.state.game_state)
    def reset(self, num_players: int, seed: Optional[int] = None):
        self.state = ta.TwoPlayerState(num_players=num_players, max_turns=self.max_turns, seed=seed)
        self.state.reset(game_state={"board": self._initialize_board()}, player_prompt_function=self._prompt, role_mapping={0:"Red", 1:"Black"})
        self.state.add_observation(message=f"Current board:\n{self._render_board()}", observation_type=ta.ObservationType.GAME_BOARD)

    def _initialize_board(self) -> List[List[str]]: return [['b' if row < 3 and (row + col) % 2 == 1 else 'r' if row > 4 and (row + col) % 2 == 1 else '.' for col in range(8)] for row in range(8)]
    def _render_board(self) -> str:
        header = "\n     " + "  ".join(str(col) for col in range(8)) + "\n"
        divider = "   +" + "-" * 25 + "\n"
        rows = "\n".join(f" {row} |" + "".join(f" {self.state.game_state["board"][row][col]} " for col in range(8)) for row in range(8))
        return header + divider + rows + "\n"

    def _prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        return (
            f"You are Player {player_id} playing a game of Checkers as {'Red' if player_id==0 else 'Black'}.\n"
            "Make your move in the format [rowFrom colFrom rowTo colTo], e.g. '[2 1 3 2]'.\nBasic rules:\n"
            "  • Move diagonally forward by 1 if empty.\n"
            "  • Capture by jumping over an opponent piece.\n"
            "  • A piece is Kinged if it reaches the opposite end.\n"
        )

    def step(self, action: str) -> Tuple[bool, ta.Info]:
        self.state.add_observation(from_id=self.state.current_player_id, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)
        self._execute_player_move(action)
        self._check_gameover()
        self.state.add_observation(message=f"Current board:\n{self._render_board()}", observation_type=ta.ObservationType.GAME_BOARD)
        return self.state.step()

    def _execute_player_move(self, action: str):
        """ Parse the action to find the requested move. If valid, make the move, otherwise set it as an invalid move """
        match = re.compile(r"\[\s*(\d)\s+(\d)\s+(\d)\s+(\d)\s*\]").search(action.strip())
        if not match: self.state.set_invalid_move(reason="No valid move format found."); return
        row_from, col_from, row_to, col_to = map(int, match.groups()) # Extract coordinates
        if self._is_valid_move(self.state.current_player_id, row_from, col_from, row_to, col_to): self._move_piece(self.state.current_player_id, row_from, col_from, row_to, col_to)
        else: self.state.set_invalid_move(reason=f"Move [{row_from} {col_from} {row_to} {col_to}] is illegal.")

    def _is_valid_move(self, player_id: int, r1: int, c1: int, r2: int, c2: int) -> bool:
        """Check if a move is valid under simplified Checkers rules. TODO: This does not handle forced captures or multi-jumps """
        if not (0 <= r1 < 8 and 0 <= c1 < 8 and 0 <= r2 < 8 and 0 <= c2 < 8):
            return False  # Out of board bounds
        piece = self.state.game_state['board'][r1][c1]; target = self.state.game_state['board'][r2][c2]
        if player_id == 0: # Check piece ownership
            if piece not in ['r', 'R']: return False # Player 0 -> must move 'r' or 'R'
        else:
            if piece not in ['b', 'B']: return False # Player 1 -> must move 'b' or 'B'
        
        if target != '.': return False   # destination must be empty
        # Simple move logic (no forced capture / multi-jump):
        dr = r2 - r1; dc = abs(c2 - c1)
        if piece in ['R', 'B']: # If it's a King, can move diagonally up or down by 1 step
            if abs(dr) == 1 and dc == 1: return True
            if abs(dr) == 2 and dc == 2: return self._is_valid_capture(r1, c1, r2, c2) # Or single capture: move by 2 if jumping over opponent
            return False
        else: # Non-king logic: must move forward 1 or capture forward
            direction = -1 if piece == 'r' else 1  # Red moves "up", Black moves "down"
            if dr == direction and dc == 1: return True # Non-capturing move
            if dr == 2 * direction and dc == 2: return self._is_valid_capture(r1, c1, r2, c2) # Capturing move
            return False

    def _is_valid_capture(self, r1: int, c1: int, r2: int, c2: int) -> bool:
        mid_r = (r1 + r2) // 2
        mid_c = (c1 + c2) // 2
        moving_piece = self.state.game_state["board"][r1][c1]
        jumped_piece = self.state.game_state["board"][mid_r][mid_c]
        if moving_piece.lower() == 'r' and jumped_piece.lower() == 'b': return True
        if moving_piece.lower() == 'b' and jumped_piece.lower() == 'r': return True
        return False

    def _move_piece(self, player_id: int, r1: int, c1: int, r2: int, c2: int):
        piece = self.state.game_state["board"][r1][c1]
        self.state.game_state["board"][r1][c1] = '.'
        self.state.game_state["board"][r2][c2] = piece

        # If capturing, remove the jumped piece
        if abs(r2 - r1) == 2:
            mid_r = (r1 + r2) // 2
            mid_c = (c1 + c2) // 2
            self.state.game_state["board"][mid_r][mid_c] = '.'

        # Check for kinging
        if piece == 'r' and r2 == 0:    self.state.game_state["board"][r2][c2] = 'R'  # Red becomes King
        elif piece == 'b' and r2 == 7:  self.state.game_state["board"][r2][c2] = 'B'  # Black becomes King
        self.state.add_observation(message=f"Player {player_id} moved ({r1},{c1}) -> ({r2},{c2}).", observation_type=ta.ObservationType.GAME_ACTION_DESCRIPTION) # Send a summary message

    def _check_gameover(self):
        red_pieces = sum(cell.lower() == 'r' for row in self.state.game_state["board"] for cell in row)
        black_pieces = sum(cell.lower() == 'b' for row in self.state.game_state["board"] for cell in row)
        if red_pieces == 0: self.state.set_winner(player_id=1, reason="Red has no pieces left. Black wins!"); return
        if black_pieces == 0: self.state.set_winner(player_id=0, reason="Black has no pieces left. Red wins!"); return
        # If either player has no legal moves, that player loses.
        if not self._has_legal_move(self.state.current_player_id): # The other player wins
            self.state.set_winners(player_id=1-self.state.current_player_id, reason=f"Player {self.state.current_player_id} has no moves left."); return
        if self.state.check_turn_limit(): self.state.set_draw(reason="The turn limit has been reached.")

    def _has_legal_move(self, player_id: int) -> bool:
        for r in range(8):
            for c in range(8):
                piece = self.state.game_state['board'][r][c]
                if player_id == 0 and piece in ['r', 'R']:
                    if self._can_piece_move(r, c): return True
                elif player_id == 1 and piece in ['b', 'B']:
                    if self._can_piece_move(r, c): return True
        return False

    def _can_piece_move(self, r: int, c: int) -> bool:
        piece = self.state.game_state['board'][r][c]
        if piece == '.': return False
        if piece in ['R', 'B']: directions = [(-1, -1), (-1, 1), (1, -1), (1, 1)] # King can move in any diagonal direction
        else:                   directions = [(-1, -1), (-1, 1)] if piece=="r" else [(1, -1), (1, 1)] # Normal piece moves (Red goes "up", Black goes "down")
        # Check single-step or capture possibility
        for dr, dc in directions:
            nr, nc = r + dr, c + dc
            if 0 <= nr < 8 and 0 <= nc < 8 and self.state.game_state['board'][nr][nc] == '.': return True
            nr2, nc2 = r + 2*dr, c + 2*dc # Check capture possibility
            if 0 <= nr2 < 8 and 0 <= nc2 < 8 and self.state.game_state['board'][nr2][nc2] == '.':
                jumped = self.state.game_state['board'][r + dr][c + dc]
                if piece.lower() == 'r' and jumped.lower() == 'b': return True
                if piece.lower() == 'b' and jumped.lower() == 'r': return True
        return False



""" Root __init__ of textarena """

from textarena.core import Env, Wrapper, ObservationWrapper, RenderWrapper, ActionWrapper, Agent, AgentWrapper, State, Message, Observations, Rewards, Info, GAME_ID, ObservationType
from textarena.state import SinglePlayerState, TwoPlayerState, FFAMultiPlayerState, TeamMultiPlayerState
from textarena.envs.registration import make, register, pprint_registry_detailed, check_env_exists
from textarena.api import make_online, make_mgc_online
from textarena import wrappers, agents

import textarena.envs

__all__ = [
    "Env", "Wrapper", "ObservationWrapper", "RenderWrapper", "ActionWrapper", "AgentWrapper", 'ObservationType', # core
    "SinglePlayerState", "TwoPlayerState", "FFAMultiPlayerState", "TeamMultiPlayerState", # state
    "make", "register", "pprint_registry_detailed", "check_env_exists", # registration
    "envs", "utils", "wrappers", # module folders
    "make_online", # play online
    "make_mgc_online", # play online with MGC
]

__version__ = "0.6.9"




from typing import Optional 

def create_board_str(board, player_id: Optional[int]=None) -> str:
    def cell_str(value: str) -> str:        return "   " if value == "." else f" {value} "
    def top_border(cols: int) -> str:       return "┌" + "───┬" * (cols - 1) + "───┐"
    def mid_border(cols: int) -> str:       return "├" + "───┼" * (cols - 1) + "───┤"
    def bottom_border(cols: int) -> str:    return "└" + "───┴" * (cols - 1) + "───┘"
    lines = []
    lines.append("  "+"   ".join(str(col) for col in range(len(board[0]))))
    lines.append(top_border(len(board[0])))
    for r in range(len(board)):
        row_line = "│" + "│".join(cell_str(val) for val in board[r]) + "│"
        lines.append(row_line)
        if r < len(board)-1:    lines.append(mid_border(len(board[0])))
        else:                   lines.append(bottom_border(len(board[0])))
    return "\n".join(lines)



import re
from typing import Optional, Dict, Tuple, Any

import textarena as ta
from textarena.envs.WildTicTacToe.renderer import create_board_str

class WildTicTacToeEnv(ta.Env):
    def __init__(self):
        super().__init__()
        self.cell_mapping = {i * 3 + j: (i, j) for i in range(3) for j in range(3)}

    def get_board_str(self): return create_board_str(board=self.state.game_state["board"])
    def _render_board(self): return "\n---+---+---\n".join("|".join(f" {self.state.game_state['board'][r][c]} " if self.state.game_state['board'][r][c] else f" {str(r * 3 + c)} " for c in range(3)) for r in range(3))
    def reset(self, num_players: int, seed: Optional[int] = None):
        self.state = ta.TwoPlayerState(num_players=num_players, seed=seed)
        self.state.reset(game_state={"board": [['' for _ in range(3)] for _ in range(3)]}, player_prompt_function=self._prompt)
        self._observer_current_state()

    def _prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        return (
            f"You are Player {player_id} in Wild Tic Tac Toe.\nOn your turn, you can place either an 'X' or an 'O' in an empty square.\n"
            "You win by aligning three of the same mark (X or O) in a row.\nYou can win with either symbol.\n"
            "Choose your move using the format '[X 4]' to place X in the center.\n"
        )

    def _observer_current_state(self):
        available_moves = [f"[{mark} {str(r*3+c)}]" for r in range(3) for c in range(3) if self.state.game_state["board"][r][c] == '' for mark in ['X', 'O']]
        self.state.add_observation(message=f"Current Board:\n\n{self._render_board()}\n\nAvailable Moves: {', '.join(available_moves)}", observation_type=ta.ObservationType.GAME_BOARD)

    def step(self, action: str) -> Tuple[bool, ta.Info]:
        self.state.add_observation(from_id=self.state.current_player_id, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)
        match = re.compile(r"\[\s*([XO])\s+(\d+)\s*\]", re.IGNORECASE).search(action)

        if match is None: self.state.set_invalid_move(reason=f"Invalid move format. Use '[X 4]' or '[O 2]' format.")
        else:
            mark = match.group(1).upper()
            cell = int(match.group(2))
            if cell not in self.cell_mapping: self.state.set_invalid_move(reason=f"Invalid cell number: {cell}. Must be between 0 and 8.")
            else:
                row, col = self.cell_mapping[cell]
                if self.state.game_state["board"][row][col] == '':
                    self.state.game_state["board"][row][col] = mark
                    self.state.add_observation(message=f"Player {self.state.current_player_id} placed their symbol ({mark}) in cell {cell}", observation_type=ta.ObservationType.GAME_ACTION_DESCRIPTION)
                    if self._check_winner(mark): self.state.set_winner(player_id=self.state.current_player_id, reason=f"Player {self.state.current_player_id} wins with {mark}s!")
                    elif all(cell != '' for row in self.state.game_state["board"] for cell in row): self.state.set_draw(reason="The game is a draw!")
                else: self.state.set_invalid_move(reason=f"Invalid move. Cell {cell} is already occupied.")
        self._observer_current_state()
        return self.state.step()

    def _check_winner(self, mark: str) -> bool:
        board = self.state.game_state["board"]
        for i in range(3):
            if all(board[i][j] == mark for j in range(3)) or all(board[j][i] == mark for j in range(3)):    return True
        if all(board[i][i] == mark for i in range(3)) or all(board[i][2 - i] == mark for i in range(3)):    return True
        return False



def create_board_str(game_state: dict) -> str:
    board = game_state["board"]

    # Mapping pieces to symbols
    symbol_map = {
        "r": "🔴", "R": "🟥",
        "b": "⚫", "B": "⬛",
        ".": "  "
    }

    lines = []
    
    # Column header
    header = "    " + "    ".join(str(c) for c in range(8))
    lines.append(header)
    
    # Top border
    lines.append("  ┌" + "────┬" * 7 + "────┐")

    for r in range(8):
        row = board[r]
        row_symbols = " │ ".join(symbol_map.get(cell, cell) for cell in row)
        lines.append(f"{r} │ {row_symbols} │")
        if r < 7:
            lines.append("  ├" + "────┼" * 7 + "────┤")
        else:
            lines.append("  └" + "────┴" * 7 + "────┘")

    return "\n".join(lines)



import re, os, random, json, importlib
from typing import Any, Dict, Optional, Tuple

import textarena as ta

class GuessWhoEnv(ta.Env):

    def __init__(self, max_turns: int=40):
        super().__init__()
        self.max_turns = max_turns
        # Initialize the gamemaster
        self.gamemaster = ta.agents.OpenRouterAgent(model_name="openai/gpt-4o")
        self.gamemaster_options = ["Yes", "No", "I don't know"]
        self.gamemaster_context = None
        self.gamemaster_history = []
        self.characters = self._load_characters() # Load character list

    def _load_characters(self, characters_path: Optional[str] = None):
        try:
            if characters_path is not None:
                if not os.path.exists(characters_path): raise FileNotFoundError(f"Characters data file not found at: {characters_path}") # Use provided path
                with open(characters_path, 'r', encoding='utf-8') as file: characters = json.load(file)
            else:
                with importlib.resources.files('textarena.envs.GuessWho').joinpath('characters.json').open('r') as file: characters = json.load(file) # Use package resource
            if not characters: raise ValueError("Characters list is empty.")
            return characters
        except Exception as e: raise FileNotFoundError(f"Failed to load characters data: {str(e)}")

    def get_gamemaster_response(self, action: str) -> str:
        """ Get the gamemaster's response based on the provided action """
        # Validate gamemaster state
        if self.gamemaster_context is None:     raise ValueError("Gamemaster context is not set.")
        if self.gamemaster_history is None:     raise ValueError("History is not set.")
        if self.gamemaster_options is None:     raise ValueError("Gamemaster options are not set.")
        options = ", ".join(f"'{opt}'" for opt in self.gamemaster_options) # Format available response options
        history = "\n".join(f"Q: {q}\nA: {a}" for q, a in self.gamemaster_history) # Construct conversation history
        prompt = f"{self.gamemaster_context}\n{history}\n\nQ: {action}\nOptions: {options}\n\nPlease respond with the most appropriate option." # Create prompt
        response = self.gamemaster(prompt).strip() # Get response from the gamemaster agent
        if any(option.lower() in response.lower() for option in self.gamemaster_options): # Validate response
            self.gamemaster_history.append((action, response))  # Store valid responses
        else:
            response = "I'm sorry, I don't understand. Please try asking again."
            self.gamemaster_history.append((action, response))  # Log fallback response
        return response

    def reset(self, num_players: int, seed: Optional[int] = None):
        self.state = ta.SinglePlayerState(num_players=num_players, seed=seed, max_turns=self.max_turns) # Initialize the game state
        self.target_character = random.choice(self.characters) ## select a random character
        self.gamemaster_context = ( ## update the gamemaster context
            f"You are the gamemaster for the game of 'Guess Who'.\n"
            f"You will provide responses to the player's questions that guides them into guessing the target character with the following name and traits: {self.target_character}.\n"
        )
        self.state.reset(game_state={"target_character": self.target_character}, player_prompt_function=self._generate_player_prompt)

    def _generate_player_prompt(self, player_id: int, game_state: Dict[int, Any]) -> str:
        """ Generate the player prompt """
        return (
            f"You are Player {player_id}. You are playing Guess Who.\n"
            "The gamemaster has chosen one target character from the list of characters that you will be shown below.\n"
            "You have to guess the target character by asking yes-or-no questions about the target character's traits.\n"
            "You can ask questions like 'Is the character male?' or 'Does the character have a beard?'.\n"
            "You can also guess the name of the target character at any time by ensuring that you wrap their name in square brackets, e.g. [Zach].\n"
            "As you play, the history of your questions and gamemaster's responses will be displayed."
            "Here is the list of characters you can ask questions about:\n"
        ) + self._characters_to_string()
    
    def _characters_to_string(self) -> str:
        formatted_descriptions = []
        for i, char in enumerate(self.characters, start=1):
            # Format the description in a narrative style
            accessories = ", ".join(char["accessories"]) if char["accessories"] else "no accessories"
            description = (
                f"{i}. {char['name']} is a {char['age_range']} {char['gender']} with {char['hair_style']} "
                f"{char['hair_color']} hair and {char['eye_color']} eyes. {char['name']} has a {char['complexion']} complexion, "
                f"{char['skin_tone']} skin tone, and {char['smile_type']} smile. They wear {accessories}, "
                f"have {char['facial_hair']} facial hair, and their clothing style is {char['clothing_style']}. "
                f"{char['name']} has {char['hair_texture']} hair texture, {char['eyewear_style']} glasses style, "
                f"a {char['nose_shape']} nose, {char['ear_size']} ears, and {char['cheek_features']} on their cheeks."
            )
            formatted_descriptions.append(description)
        return "\n\n".join(formatted_descriptions) # Join all descriptions into a single text block
    
    def step(self, action: str) -> Tuple[bool, ta.Info]:
        """ Process the player's action and update the environment state """
        player_id = self.state.current_player_id
        self.state.add_observation(from_id=player_id, to_id=-1, message=action, observation_type=ta.ObservationType.PLAYER_ACTION) ## update the observation
        action_search_pattern = re.compile(r"\[([a-zA-Z]+)\]") # e.g. [zach]
        action_match = action_search_pattern.search(action)
        self.state.add_observation(message=f"You asked: {action}", observation_type=ta.ObservationType.GAME_ACTION_DESCRIPTION)
        if not action_match: self.state.add_observation(from_id=-1, to_id=player_id, message=self.get_gamemaster_response(action), observation_type=ta.ObservationType.GAME_MESSAGE) ## add the gamemaster response to the game state
        else: ## if the action is a guess
            action_text = action_match.group(1).lower()
            if action_text == self.target_character["name"].lower():    self.state.set_outcome(reward=1, reason=f"Congratulations! Player {player_id} guessed the target character.")
            else:                                                       self.state.set_invalid_move(reason=f"Invalid guess. Player {player_id} guessed incorrectly.")
        if self.state.check_turn_limit(): self.state.set_outcome(reward=0, reason="The turn limit has been reached.")
        return self.state.step()
    


import re, random, copy
from typing import Any, Dict, List, Tuple, Optional

import nltk
from nltk.corpus import words
nltk.download('words')

import textarena as ta
from textarena.envs.Hangman.renderer import create_board_str


class HangmanEnv(ta.Env):
    def __init__(self, hardcore: Optional[bool] = False):
        """ 
        Args:
            hardcore: Whether to play in hardcore mode.
        """
        super().__init__()
        self.hardcore = hardcore
        self.word_list = words.words("en") if hardcore else words.words("en-basic") ## load the word list (to be sampled from)

    def get_board_str(self):
        return create_board_str(game_state=self.state.game_state)

    def reset(self, num_players: int, seed: Optional[int]=None):
        self.state = ta.SinglePlayerState(num_players=num_players, seed=seed) ## initialize the game state
        target_word = random.choice(self.word_list)
        game_state = {
            "target_word": target_word, "target_letters": list(target_word.upper()), 
            "current_board": ["_" for _ in target_word], "guessed_letters": set(), "tries_left":6
        }
        self.state.reset(game_state=game_state, player_prompt_function=self._generate_player_prompt)
        self._observe_current_state()
    
    def _generate_player_prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        return (
            f"You are playing Hangman. The objective of the game is to guess the word by providing one letter guesses or the entire word.\n"
            "Each column is numbered. The cells that need to be populated with letters are represented by '_'.\n\n"
            "There are two ways you can answer. You can provide one letter guesses in the format of '[L]', or you can guess the entire word in the format of '[LIGHT]'.\n"
            "If the given letter is in the word, it will be revealed in the grid.\n"
            "If the given word is correct, you win.\n"
            "As you play, the history of your choices will be appended below. Use the information to figure out the word and win.\n"
            "You have 6 incorrect tries before the game ends.\n\n"
        )
    
    def _observe_current_state(self) -> None:
        message = f"Current board:\n{self._render_current_board()}\nYou have {self.state.game_state['tries_left']} tries left.\nGuessed letters: {', '.join(sorted(self.state.game_state['guessed_letters']))}"
        self.state.add_observation(message=message, observation_type=ta.ObservationType.GAME_BOARD)
    
    def _render_current_board(self) -> str:
        lines = [" ".join(f"C{i:02}" for i in range(len(self.state.game_state["current_board"])))]
        row_str = ""  # Label for the single row
        for i, val in enumerate(self.state.game_state["current_board"]): row_str += f"  {val} "
        lines.append(row_str)
        return "\n"+"\n".join(lines)
    
    def step(self, action: str) -> Tuple[bool, ta.Info]:
        """ Process the player's action and update the game state accordingly """
        self.state.add_observation(from_id=self.state.current_player_id, message=action, observation_type=ta.ObservationType.PLAYER_ACTION) # Update the observations
        match = re.compile(r"\[([a-zA-Z]+)\]", re.IGNORECASE).search(action)

        if not match:
            self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=f"Invalid move format. You did not respond with a valid 'letter' or 'word'.")
        else:
            # for match in matches:
            letter = match.group(1).upper()  # Convert to uppercase for consistency
            if len(letter) > 1: # Player guessed full word
                if letter == self.state.game_state["target_word"].upper():
                    self.state.set_outcome(reward=1, reason=f"Congratulations! You completed the Hangman puzzle.")
                    self.state.game_state["current_board"] = self.state.game_state["target_letters"]  # reveal the word
                else:
                    self.state.add_observation(message=f"Your guess of '{letter}' is not the target word.", observation_type=ta.ObservationType.GAME_MESSAGE)

            else: # Player guessed a single letter
                if letter in self.state.game_state["guessed_letters"]: # Check if the letter has been guessed before
                    self.state.set_invalid_move(reward=self._get_percentage_completion(), reason=f"You guessed the letter '{letter}' which has already been guessed.") 
                else:
                    self.state.game_state["guessed_letters"].add(letter)
                    if letter in self.state.game_state["target_letters"]: # Check if the letter is in the target word
                        self._reveal_letter(letter) # Update the word progress to reveal this letter
                        self.state.add_observation(message=f"Your guess of {letter} is in the word", observation_type=ta.ObservationType.GAME_MESSAGE)
                    else:
                        self.state.game_state["tries_left"] -= 1
                        self.state.add_observation(message=f"Your guess of {letter} is not in the word. You have {self.state.game_state['tries_left']} lives left.", observation_type=ta.ObservationType.GAME_MESSAGE)
                    self.state.add_observation(self._render_current_board(), observation_type=ta.ObservationType.GAME_BOARD)

            if self.state.game_state["tries_left"] == 0:                                            self.state.set_outcome(reward=self._get_percentage_completion(), reason=f"You are out of tries. You guessed {self._get_percentage_completion()*100:.2f} percentage of the characters correctly. The target word was : {self.state.game_state['target_word']}")
            elif self.state.game_state["current_board"] == self.state.game_state["target_letters"]: self.state.set_outcome(reward=1, reason=f"Congratulations! You have completed the Hangman puzzle.")
        return self.state.step()
    
    def _reveal_letter(self, letter: str) -> None:
        for i, char in enumerate(self.state.game_state["target_letters"]):
            if char == letter: self.state.game_state["current_board"][i] = letter

    def _get_percentage_completion(self) -> float:
        return sum(1 for a, b in zip(self.state.game_state["current_board"], self.state.game_state["target_word"]) if a.upper() == b.upper()) / len(self.state.game_state["target_word"])

    



from enum import Enum
import re, random
from typing import Tuple, Dict, Optional, List
import textarena as ta

class Phase(Enum):
    NIGHT_MAFIA = "Night-Mafia"
    NIGHT_DOCTOR = "Night-Doctor"
    NIGHT_DETECTIVE = "Night-Detective"
    DAY_DISCUSSION = "Day-Discussion"
    DAY_VOTING = "Day-Voting"

class Role:
    name: str = "Role"
    team: str = "Unknown"
    description: str = ""
    def get_prompt(self, player_id: int, player_roles: Dict[int, str], num_players: int, num_discussion_rounds: int) -> str: raise NotImplementedError

class Villager(Role):
    name = "Villager"
    team = "Village"
    description = "A regular villager. Your goal is to identify and eliminate all Mafia members through voting during the day."
    def get_prompt(self, player_id, player_roles, num_players, num_discussion_rounds):
        return (
            f"Welcome to Secret Mafia! You are Player {player_id}.\n"
            f"Your role: {self.name}\nTeam: {self.team}\nDescription: {self.description}\n\n"
            f"Players: {', '.join([f'Player {i}' for i in range(num_players)])}\n\n"
            f"The game progresses through Day and Night phases.\n"
            f"- During the Day phase, there are {num_discussion_rounds} rounds of discussion followed by voting.\n"
            f"- During discussions, everything you say is automatically broadcasted to all players.\n"
            f"- After discussions, all players must vote to eliminate one player.\n"
            f"- During the Night phase, you have no special actions.\n\n"
            f"The game ends when either all Mafia members are eliminated (Village wins) or\n"
            f"Mafia members equal or outnumber Villagers (Mafia wins).\n"
        )

class Mafia(Role):
    name = "Mafia"
    team = "Mafia"
    description = "A Mafia member. Eliminate villagers and gain majority."
    def get_prompt(self, player_id, player_roles, num_players, num_discussion_rounds):
        teammates = [f"Player {pid}" for pid, r in player_roles.items() if r == "Mafia"]
        return (
            f"Welcome to Secret Mafia! You are Player {player_id}.\n"
            f"Your role: {self.name}\nTeam: {self.team}\nDescription: {self.description}\n\n"
            f"Players: {', '.join([f'Player {i}' for i in range(num_players)])}\n\n"
            f"Your teammates are: {', '.join(teammates)}.\n\n"
            f"During DAY phase: Speak freely and vote.\n"
            f"During NIGHT phase: '[Player X]' to vote and eliminate a villager.\n"
            f"Win by eliminating villagers until Mafia equal or outnumber them.\n"
        )

class Doctor(Role):
    name = "Doctor"
    team = "Village"
    description = "Protect one player each night from Mafia elimination."
    def get_prompt(self, player_id, player_roles, num_players, num_discussion_rounds):
        return (
            f"Welcome to Secret Mafia! You are Player {player_id}.\n"
            f"Your role: {self.name}\nTeam: {self.team}\nDescription: {self.description}\n\n"
            f"Players: {', '.join([f'Player {i}' for i in range(num_players)])}\n\n"
            f"During DAY phase: Speak freely and vote.\n"
            f"During NIGHT phase: '[Player X]' to protect a player.\n"
            f"Win by identifying and eliminating all Mafia members.\n"
        )

class Detective(Role):
    name = "Detective"
    team = "Village"
    description = "Investigate players to find Mafia members."
    def get_prompt(self, player_id, player_roles, num_players, num_discussion_rounds):
        return (
            f"Welcome to Secret Mafia! You are Player {player_id}.\n"
            f"Your role: {self.name}\nTeam: {self.team}\nDescription: {self.description}\n\n"
            f"Players: {', '.join([f'Player {i}' for i in range(num_players)])}\n\n"
            f"During DAY phase: Speak freely and vote.\n"
            f"During NIGHT phase: '[Player X]' to investigate.\n"
            f"You'll learn immediately if the target is Mafia.\n"
            f"Win by identifying and eliminating all Mafia members.\n"
        )

class VoteHandler:
    @staticmethod
    def parse(text: str) -> Optional[int]:
        m = SecretMafiaEnv.voting_pattern.search(text)
        return int(m.group(1)) if m else None
    @staticmethod
    def tally(votes: Dict[int, int]) -> Optional[int]:
        if not votes: return None
        # Count votes per target
        counts: Dict[int, int] = {}
        for target in votes.values():
            counts[target] = counts.get(target, 0) + 1
        top_score = max(counts.values()) # Highest vote count
        top_players = [pid for pid, c in counts.items() if c == top_score] # All players who received the top score (could be 1 or many)
        return random.choice(top_players) # Randomly resolve ties

class SecretMafiaEnv(ta.Env):
    voting_pattern = re.compile(r".*\[(?:player\s*)?(\d+)\].*", re.IGNORECASE)
    _ROLE_FACTORY = {
        "Villager":  Villager,
        "Mafia":     Mafia,
        "Doctor":    Doctor,
        "Detective": Detective,
    }
    def __init__(self, mafia_ratio: float = 0.25, discussion_rounds: int = 3):
        """
        Args:
            mafia_ratio (float): Ratio of Mafia members to total players (default: 0.25)
            discussion_rounds (int): The number of discussion rounds
        """
        self.mafia_ratio = mafia_ratio
        self.discussion_rounds = discussion_rounds

    def reset(self, num_players: int, seed: Optional[int] = None):
        assert 6 <= num_players <= 15, "Player count must be between 5 and 15."
        self.state = ta.TeamMultiPlayerState(num_players=num_players, seed=seed)
        self._assign_roles(num_players)
        self.phase: Phase = Phase.NIGHT_MAFIA
        game_state = {
            "phase": self.phase,
            "day_number": 1,
            "alive_players": list(range(num_players)),
            "player_roles": self.player_roles,
            "num_discussion_rounds": self.discussion_rounds,
            "votes": {},
            "pending_elimination": None,
        }
        self.state.reset(game_state=game_state, player_prompt_function=self._prompt, secret_roles=self.player_roles)
        self._send_phase_prompts() # populate self.next_player_ids
        self.state.manually_set_current_player_id(self.next_player_ids.pop())
    

    def _assign_roles(self, num_players: int):
        self.player_roles = {}
        self.roles = {}                              # <- NEW
        num_mafia = max(1, round(num_players * self.mafia_ratio))
        role_pool = ["Mafia"] * num_mafia + ["Doctor", "Detective"] 
        role_pool += ["Villager"] * (num_players - len(role_pool))
        random.shuffle(role_pool)

        for pid, r_name in enumerate(role_pool):
            self.player_roles[pid] = r_name
            self.roles[pid] = self._ROLE_FACTORY[r_name]()

    def _prompt(self, player_id: int, game_state: dict) -> str:
        role_obj = self.roles[player_id]
        return role_obj.get_prompt(player_id = player_id, player_roles = self.player_roles, num_players = self.state.num_players, num_discussion_rounds = self.discussion_rounds)

    def step(self, action: str) -> Tuple[bool, ta.Info]:
        pid = self.state.current_player_id
        phase_dispatch = {
            Phase.DAY_DISCUSSION: self._handle_discussion, Phase.DAY_VOTING: self._handle_day_vote, Phase.NIGHT_MAFIA: self._handle_mafia_vote, 
            Phase.NIGHT_DOCTOR: self._handle_doctor_action, Phase.NIGHT_DETECTIVE: self._handle_detective_action,
        }
        phase_dispatch[self.phase](pid, action)
        self._after_player_action() # rotate / advance phase
        return self.state.step(rotate_player=False)

    def _after_player_action(self):
        if self.state.made_invalid_move: return
        # If players still queued, just rotate.
        if self.next_player_ids:
            self.state.manually_set_current_player_id(self.next_player_ids.pop())
            return

        # Phase complete ─ evaluate votes / killings, decide next phase, queue players
        if self.phase == Phase.DAY_VOTING:      self._resolve_day_votes()
        elif self.phase == Phase.NIGHT_MAFIA:   self._store_mafia_target()

        # When night sequence ends (Doctor or Detective → Day)
        if self.phase in (Phase.NIGHT_DOCTOR, Phase.NIGHT_DETECTIVE):
            next_phase = self._compute_next_phase()
            if next_phase == Phase.DAY_DISCUSSION:
                self._resolve_night_outcome()

        # Advance to next phase
        self.phase = self._compute_next_phase()
        self.state.game_state["phase"] = self.phase
        self._send_phase_prompts()
        self.state.manually_set_current_player_id(self.next_player_ids.pop())

    def _compute_next_phase(self) -> Phase:
        doctor_alive     = any(self.player_roles[p] == "Doctor"    for p in self.state.game_state["alive_players"])
        detective_alive  = any(self.player_roles[p] == "Detective" for p in self.state.game_state["alive_players"])
        match self.phase:
            case Phase.NIGHT_MAFIA:     return Phase.NIGHT_DOCTOR if doctor_alive else (Phase.NIGHT_DETECTIVE if detective_alive else Phase.DAY_DISCUSSION)
            case Phase.NIGHT_DOCTOR:    return Phase.NIGHT_DETECTIVE if detective_alive else Phase.DAY_DISCUSSION
            case Phase.NIGHT_DETECTIVE: return Phase.DAY_DISCUSSION
            case Phase.DAY_DISCUSSION:  return Phase.DAY_VOTING
            case Phase.DAY_VOTING:      return Phase.NIGHT_MAFIA
            case _:                     raise RuntimeError("Unknown phase")
                

    def _send_phase_prompts(self):
        gs = self.state.game_state
        alive = gs["alive_players"]
        self.next_player_ids: List[int] = []

        if self.phase == Phase.NIGHT_MAFIA:
            mafia = [p for p in alive if self.player_roles[p] == "Mafia"]
            targets = [p for p in alive if p not in mafia]
            for p in mafia:
                self.state.add_observation(to_id=p, message=f"Night has fallen. Mafia, agree on a victim.\nValid targets: {', '.join(f'[{t}]' for t in targets)}", observation_type=ta.ObservationType.GAME_MESSAGE)
            self.next_player_ids = random.sample(mafia, k=len(mafia))

        elif self.phase == Phase.NIGHT_DOCTOR:
            doc = next(p for p in alive if self.player_roles[p] == "Doctor")
            opts = ", ".join(f"[{t}]" for t in alive if t != doc)
            self.state.add_observation(to_id=doc, message=f"Night phase - choose one player to protect: {opts}", observation_type=ta.ObservationType.GAME_MESSAGE)
            self.next_player_ids = [doc]

        elif self.phase == Phase.NIGHT_DETECTIVE:
            det = next(p for p in alive if self.player_roles[p] == "Detective")
            opts = ", ".join(f"[{t}]" for t in alive if t != det)
            self.state.add_observation(to_id=det, message=f"Night phase - choose one player to investigate: {opts}", observation_type=ta.ObservationType.GAME_MESSAGE)
            self.next_player_ids = [det]

        elif self.phase == Phase.DAY_DISCUSSION:
            rounds = self.discussion_rounds
            self.state.add_observation(to_id=-1, message=f"Day breaks. Discuss for {rounds} rounds, then a vote will follow.", observation_type=ta.ObservationType.GAME_MESSAGE)
            players = random.sample(alive, k=len(alive))
            self.next_player_ids = players * rounds

        elif self.phase == Phase.DAY_VOTING:
            opts = ", ".join(f"[{p}]" for p in alive)
            self.state.add_observation(to_id=-1, message=f"Voting phase - submit one vote in format [X]. Valid: {opts}", observation_type=ta.ObservationType.GAME_MESSAGE)
            self.next_player_ids = random.sample(alive, k=len(alive))

    def _handle_discussion(self, pid: int, action: str):    self.state.add_observation(from_id=pid, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)
    def _handle_day_vote(self, pid: int, action: str):      self._record_vote(pid, action, broadcast_to_all=True)
    def _handle_mafia_vote(self, pid: int, action: str):    self._record_vote(pid, action, broadcast_to_mafia_only=True)
    def _handle_doctor_action(self, pid: int, action: str):
        target = VoteHandler.parse(action)
        if target is None or target not in self.state.game_state["alive_players"]:
            fatal = self._mark_invalid(pid, "Invalid protection target.")
            if not fatal: 
                return
            else: # player was eliminated by invalid move
                self.state.made_invalid_move = False  # such that we can rotate off the player 
                return

        # save target
        if target == self.state.game_state["pending_elimination"]:
            self.state.game_state["pending_elimination"] = None
        self.state.add_observation(from_id=pid, to_id=pid, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)

    def _handle_detective_action(self, pid: int, action: str):
        target = VoteHandler.parse(action)
        if target is None or target not in self.state.game_state["alive_players"]:
            fatal = self._mark_invalid(pid, "Invalid investigation target.")
            if not fatal: return
            else:# player was eliminated by invalid move
                self.state.made_invalid_move = False  # such that we can rotate off the player 
                return
        is_mafia = self.player_roles[target] == "Mafia"
        result = f"Player {target} IS{' ' if is_mafia else ' NOT '}a Mafia member."
        self.state.add_observation(to_id=pid, message=result, observation_type=ta.ObservationType.GAME_MESSAGE)

    def _record_vote(self, pid: int, action: str, *, broadcast_to_all=False, broadcast_to_mafia_only=False):
        target = VoteHandler.parse(action)
        if target is None or target not in self.state.game_state["alive_players"]:
            fatal = self._mark_invalid(pid, "Vote not in valid format or invalid target.")
            if not fatal: return
            else: # player was eliminated by invalid move
                self.state.made_invalid_move = False  # such that we can rotate off the player 
                return


        self.state.game_state["votes"][pid] = target

        if broadcast_to_all:
            self.state.add_observation(from_id=pid, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)
        elif broadcast_to_mafia_only:
            mafia = [p for p in self.state.game_state["alive_players"] if self.player_roles[p] == "Mafia"]
            for m in mafia:
                self.state.add_observation(from_id=pid, to_id=m, message=action, observation_type=ta.ObservationType.PLAYER_ACTION)

    def _mark_invalid(self, pid: int, reason: str):
        fatal = self.state.set_invalid_move(reason)
        if fatal: self._eliminate_player(self.state.current_player_id, "has been eliminated by making an invalid move.")
        return fatal 
    
        # if self.state.set_invalid_move(reason):
        #     # repeated invalid move by player, kill'em
        #     self._eliminate_player(self.state.current_player_id, "has been eliminated by making an invalid move.")

            # # TODO kill player off
            # others = [p for p in range(self.state.num_players) if p != pid]
            # self.state.set_winners(player_ids=others, reason=f"Player {pid} made an invalid move.")

    def _resolve_day_votes(self):
        target = VoteHandler.tally(self.state.game_state["votes"])
        self.state.game_state["votes"].clear()
        if target is None:
            self.state.add_observation(message="No consensus - nobody was eliminated.", observation_type=ta.ObservationType.GAME_MESSAGE)
            return
        self._eliminate_player(target, "was eliminated by vote")

    def _store_mafia_target(self):
        self.state.game_state["pending_elimination"] = VoteHandler.tally(self.state.game_state["votes"])
        self.state.game_state["votes"].clear()

    def _resolve_night_outcome(self):
        tgt = self.state.game_state["pending_elimination"]
        self.state.game_state["pending_elimination"] = None
        if tgt is None:
            self.state.add_observation(message="No one was killed tonight.", observation_type=ta.ObservationType.GAME_MESSAGE)
        else:
            self._eliminate_player(tgt, "was killed during the night")

    def _eliminate_player(self, pid: int, reason: str):
        if pid in self.state.game_state["alive_players"]:
            self.state.game_state["alive_players"].remove(pid)
        self.state.add_observation(message=f"Player {pid} {reason}.", observation_type=ta.ObservationType.GAME_MESSAGE)
        self._check_win()

    def _check_win(self):
        alive = self.state.game_state["alive_players"]
        mafia_alive = [p for p in alive if self.player_roles[p] == "Mafia"]

        if not mafia_alive:
            villagers = [p for p in range(self.state.num_players) if self.player_roles[p] != "Mafia"]
            self.state.set_winners(player_ids=villagers, reason="All Mafia were eliminated. Village wins!")
        elif len(mafia_alive) >= len(alive) / 2:
            mafia = [p for p in range(self.state.num_players) if self.player_roles[p] == "Mafia"]
            self.state.set_winners(player_ids=mafia, reason="Mafia reached parity with villagers. Mafia wins!")



import re, random
from typing import Any, Dict, Optional, Tuple, List

import textarena as ta


class NegotiationEnv(ta.Env):
    """
    N-player Negotiation Game with the ability to:
      - broadcast messages,
      - send private messages,
      - make multiple offers to specific opponents,
      - accept/deny multiple pending offers.
    """

    # --- Updated Regex Patterns ---
    # Broadcast: three alternatives to capture:
    #   A: [Broadcast: message] inside the brackets
    #   B: [Broadcast message] inside the brackets (no colon)
    #   C: [Broadcast] message (outside the brackets)
    broadcast_pattern = re.compile(
        r"(?:"
        r"\s*\[Broadcast\s*:\s*(.*?)\]"         # Alternative A: colon present.
        r"|"
        r"\s*\[Broadcast((?:\s+).*?)\]"           # Alternative B: no colon, whitespace inside.
        r"|"
        r"\s*\[Broadcast\](\s+.*?)(?=\s*\[|$)"    # Alternative C: message appears after the bracket.
        r")",
        re.IGNORECASE | re.DOTALL
    )

    # Whisper: require a player id and a colon
    whisper_pattern = re.compile(
        r"\s*\[Whisper\s+(?:to\s+)?(?:Player\s+)?(\d+)\s*:\s*(.*?)\]",
        re.IGNORECASE | re.DOTALL
    )

    # The rest remain unchanged
    offer_pattern = re.compile(
        r"\[Offer\s+(?:to\s+)?(?:Player\s+)?(\d+)\s*:?\s*(.*?)\]",
        re.IGNORECASE | re.DOTALL
    )
    accept_pattern = re.compile(r"\[Accept\s*#?\s*(\d+)\]", re.IGNORECASE)
    deny_pattern = re.compile(r"\[Deny\s*#?\s*(\d+)\]", re.IGNORECASE)

    def __init__(self, turn_multiple: int = 3):
        """
        Initialize the N-player Negotiation Game environment.

        Args:
            turn_multiple (int): Number of turns per player
        """
        
        self.resource_names = ["Wheat", "Wood", "Sheep", "Brick", "Ore"]
        self.base_values = {"Wheat": 5, "Wood": 10, "Sheep": 15, "Brick": 25, "Ore": 40}

        self.turn_multiple = turn_multiple

    @property
    def terminal_render_keys(self):
        return ["player_resources", "player_values", "pending_offers"]

    def reset(self, num_players: int, seed: Optional[int] = None):
        """ Reset the environment to its initial state """
        # Create the underlying game state
        self.state = ta.State(num_players=num_players, min_players=2, max_players=15, max_turns=int(num_players*self.turn_multiple), seed=seed)

        # Initialize each player's resources to random amounts
        # and each player's private values for resources
        player_resources = {}
        player_values = {}
        for pid in range(self.state.num_players):
            # random resource counts
            player_resources[pid] = {
                r: random.randint(5, 25) for r in self.resource_names
            }
            # random personal valuations (±20% around base, but clamp 5..40)
            player_values[pid] = {}
            for r in self.resource_names:
                base = self.base_values[r]
                variation = int(0.2 * base)  # ±20%
                low, high = max(5, base - variation), min(40, base + variation)
                player_values[pid][r] = random.randint(low, high)

        # Reset the State
        game_state = {
            "player_resources": player_resources,
            "player_values": player_values,
            "pending_offers": {},
            "offer_id_counter": 0,
        }
        self.state.reset(game_state=game_state, player_prompt_function=self._generate_player_prompt)


    def _generate_player_prompt(self, player_id: int, game_state: Dict[str, Any]) -> str:
        """
        Generate the prompt for each player on reset (or new round).
        Display their resource counts and personal valuations.
        """
        resources = game_state["player_resources"][player_id]
        valuations = game_state["player_values"][player_id]
        resource_lines = []
        for r in self.resource_names:
            qty = resources[r]
            val = valuations[r]
            resource_lines.append(f"- {qty} x {r} (value: {val} each)")
        resource_str = "\n".join(resource_lines)

        prompt = (
            f"You are Player {player_id} in a multi-player game of Negotiation with {self.state.num_players} players.\n"
            f"You have:\n{resource_str}\n\n"
            "You can broadcast messages, privately message someone, or make trade offers.\n"
            "You can also accept or deny any offers you received previously.\n"
            f"Your personal valuations are shown above; your goal is to maximize your total resource value.\n"
            "Available actions:\n"
            "  '[Broadcast: Some message]' - Send a message to all players\n"
            "  '[Whisper to X: Some message]' - Send a private message to a specific player\n"
            "  '[Offer to X: 2 Wheat -> 3 Wood]' - Make a trade offer to a specific player\n"
            "  '[Accept <x>]' or '[Deny <x>]' - Accept or Deny a trade offer\n"
            "You may combine multiple tokens in a single turn if you like.\n"
        )
        if self.state.max_turns is not None:
            prompt += f"Game ends after {self.state.max_turns} turns.\n"
        return prompt

    def step(self, action: str) -> Tuple[bool, ta.Info]:
        """
        Process a single step from the current player. That action may
        include multiple tokens (broadcast, message, offers, accepts, denies).
        """
        current_pid = self.state.current_player_id
        # Log the raw action for debugging
        # self.state.add_log(from_id=current_pid, message=action)
        self.state.add_observation(from_id=current_pid, to_id=current_pid, message=action)

        # 1. Parse out all tokens in the player's action
        #    We can do each in sequence so a single "action" can contain many instructions.
        #    For a simpler approach, you might limit to a single instruction per turn.
        self._process_broadcasts(current_pid, action)
        self._process_private_messages(current_pid, action)
        self._process_offers(current_pid, action)
        self._process_accepts_and_denies(current_pid, action)

        # 2. If we've reached turn limit, optionally decide a winner or call it a draw
        if self.state.turn == (self.state.max_turns or 0) - 1:
            self._determine_winner()

        # 3. End the step: rotate to next player
        return self.state.step()

    # --- New Helper Methods for Broadcasts and Whispers ---
    def _parse_broadcast(self, text: str) -> List[str]:
        """
        Process text to extract broadcast messages.
        For each match (a 3-tuple), pick the first non-empty capture.
        Ensure the returned message starts with a space.
        """
        results = []
        raw = self.broadcast_pattern.findall(text)
        for g1, g2, g3 in raw:
            msg = g1 or g2 or g3
            if msg and msg.strip():
                # Prepend a space if not present.
                if not msg.startswith(" "):
                    msg = " " + msg
                results.append(msg)
        return results

    def _parse_whisper(self, text: str) -> List[Tuple[str, str]]:
        """
        Process text to extract whisper tokens.
        Returns a list of (target_pid, message) tuples,
        ensuring the message starts with a space.
        """
        results = []
        matches = self.whisper_pattern.findall(text)
        for pid_str, msg in matches:
            if msg and not msg.startswith(" "):
                msg = " " + msg
            results.append((pid_str, msg))
        return results

    # --- Processing Methods ---
    def _process_broadcasts(self, from_pid: int, action: str):
        """Find all `[Broadcast...]` patterns and send them to everyone."""
        messages = self._parse_broadcast(action)
        for msg_content in messages:
            # Note: msg_content already has a leading space.
            if msg_content.strip():
                message = f"(Broadcast) Player {from_pid} says:{msg_content}"
                self.state.add_observation(from_id=from_pid, to_id=-1, message=message)

    def _process_private_messages(self, from_pid: int, action: str):
        """
        Find all `[Whisper to X: ...]` patterns, send them only to X.
        Everyone sees that from_pid performed an action, but not the content itself.
        """
        matches = self._parse_whisper(action)
        for target_str, msg_content in matches:
            msg_content = msg_content.strip()
            try:
                target_pid = int(target_str)
            except ValueError:
                self.state.set_invalid_move(player_id=from_pid, reason=f"Invalid private-message target: {target_str}")
                continue

            if target_pid not in range(self.state.num_players):
                self.state.set_invalid_move(player_id=from_pid, reason=f"Attempted to message a non-existent player {target_pid}.")
                continue

            if msg_content:
                message = f"(Private) Player {from_pid} says:{msg_content}"
                self.state.add_observation(from_id=from_pid, to_id=target_pid, message=message)
            else:
                self.state.set_invalid_move(player_id=from_pid, reason="Empty private message?")

    def _process_offers(self, from_pid: int, action: str):
        """
        Find all `[Offer to X: ... -> ...]` patterns.  
        Each is stored in game_state['pending_offers'] with a new ID.
        """
        game_state = self.state.game_state
        matches = self.offer_pattern.findall(action)
        # Each match returns: (target_pid_str, "2 Wood -> 1 Ore")
        for target_str, offer_str in matches:
            offer_str = offer_str.strip()
            try:
                target_pid = int(target_str)
            except ValueError:
                self.state.set_invalid_move(player_id=from_pid, reason=f"Invalid offer target: {target_str}")
                continue
            if target_pid not in range(self.state.num_players):
                self.state.set_invalid_move(player_id=from_pid, reason=f"Offer made to invalid player ID {target_pid}")
                continue

            # Split "2 Wood -> 1 Ore" using "->" as the delimiter.
            parts = re.split(r"->", offer_str)
            if len(parts) != 2:
                reason = f"Cannot parse Offer: '{offer_str}'. Must be like '2 Wheat -> 3 Wood'."
                self.state.set_invalid_move(player_id=from_pid, reason=reason)
                continue

            offered_str = parts[0].strip()
            requested_str = parts[1].strip()

            # Parse resources
            offered_dict = self._parse_resource_list(offered_str)
            requested_dict = self._parse_resource_list(requested_str)
            if offered_dict is None or requested_dict is None:
                self.state.set_invalid_move(player_id=from_pid, reason=f"Invalid resource format in offer: '{offer_str}'")
                continue

            # Check if the offering player has enough resources to cover what they are offering
            if not self._check_sufficient_resources(from_pid, offered_dict, game_state["player_resources"]):
                reason = f"You do not hold enough resources to offer {offered_dict} to Player {target_pid}."
                self.state.set_invalid_move(player_id=from_pid, reason=reason)
                continue

            # Create a new offer ID
            game_state["offer_id_counter"] += 1
            new_id = game_state["offer_id_counter"]
            game_state["pending_offers"][new_id] = {
                "from": from_pid,
                "to": target_pid,
                "offered_resources": offered_dict,
                "requested_resources": requested_dict
            }

            # Broadcast a FYI that an offer was created (without details)
            message = f"Offer #{new_id} created: Player {from_pid} -> Player {target_pid}."
            self.state.add_observation(from_id=ta.GAME_ID, to_id=-1, message=message)
            # Let target know they've received a new offer
            message = (
                f"You have a new offer [ID #{new_id}] from Player {from_pid}: "
                f"{self._offer_to_str(offered_dict, requested_dict)}\n"
                f"You can [accept #{new_id}] or [deny #{new_id}] it."
            )
            self.state.add_observation(from_id=ta.GAME_ID, to_id=target_pid, message=message)

    def _process_accepts_and_denies(self, current_pid: int, action: str):
        """
        Find `[Accept #ID]` and `[Deny #ID]`.
        - Only the 'to' player of that offer can accept/deny it.
        - On accept, verify that both sides have the resources needed to fulfill.
          If valid, exchange resources and remove the offer from pending_offers.
        - On deny, simply remove the offer from pending_offers.
        """
        accepts = self.accept_pattern.findall(action)  # list of IDs
        denies  = self.deny_pattern.findall(action)    # list of IDs

        for offer_id_str in accepts:
            try:
                offer_id = int(offer_id_str)
            except ValueError:
                self.state.set_invalid_move(player_id=current_pid, reason=f"Invalid offer ID in Accept: {offer_id_str}")
                continue
            self._attempt_accept_offer(current_pid, offer_id)

        for offer_id_str in denies:
            try:
                offer_id = int(offer_id_str)
            except ValueError:
                self.state.set_invalid_move(player_id=current_pid, reason=f"Invalid offer ID in Deny: {offer_id_str}")
                continue
            self._deny_offer(current_pid, offer_id)


    def _attempt_accept_offer(self, current_pid: int, offer_id: int):
        """
        Attempt to accept an offer from game_state['pending_offers'][offer_id].
        Confirm that `current_pid` is the 'to' player. Check resources, then exchange.
        """
        game_state = self.state.game_state
        if offer_id not in game_state["pending_offers"]:
            self.state.set_invalid_move(player_id=current_pid, reason=f"Offer #{offer_id} does not exist.")
            return
        off = game_state["pending_offers"][offer_id]
        if off["to"] != current_pid:
            self.state.set_invalid_move(player_id=current_pid, reason=f"Offer #{offer_id} is not addressed to you.")
            return

        # Check if from-player still has what they offered
        if not self._check_sufficient_resources(off["from"], off["offered_resources"], game_state["player_resources"]):
            # The offering player no longer has enough resources to complete the trade
            self.state.add_observation(message=f"Offer #{offer_id} canceled because Player {off['from']} no longer has enough resources to fulfill it.")
            del game_state["pending_offers"][offer_id]
            return

        # Check if to-player has what is requested
        if not self._check_sufficient_resources(
            off["to"], off["requested_resources"], game_state["player_resources"]
        ):
            reason=f"You do not have enough resources to fulfill Offer #{offer_id}."
            self.state.set_invalid_move(player_id=current_pid, reason=reason)
            return

        # Execute the trade
        self._exchange_resources(off["from"], off["to"],
                                 off["offered_resources"], off["requested_resources"])
        # Announce success
        message=(
            f"Player {off['to']} ACCEPTED Offer #{offer_id} from Player {off['from']}: "
            f"{self._offer_to_str(off['offered_resources'], off['requested_resources'])}"
        )
        self.state.add_observation(from_id=ta.GAME_ID, to_id=-1, message=message)
        # Remove it from pending
        del game_state["pending_offers"][offer_id]

    def _deny_offer(self, current_pid: int, offer_id: int):
        """ Deny (remove) the specified offer, if it targets the current player """
        game_state = self.state.game_state
        if offer_id not in game_state["pending_offers"]:
            self.state.set_invalid_move(player_id=current_pid, reason=f"Offer #{offer_id} does not exist.")
            return

        off = game_state["pending_offers"][offer_id]
        if off["to"] != current_pid:
            self.state.set_invalid_move(player_id=current_pid, reason=f"Offer #{offer_id} is not addressed to you.")
            return

        message=f"Player {current_pid} DENIED Offer #{offer_id} from Player {off['from']}."
        self.state.add_observation(from_id=ta.GAME_ID, to_id=-1, message=message)
        del game_state["pending_offers"][offer_id]

    def _check_sufficient_resources(self, pid: int, needed: Dict[str, int],
                                    all_resources: Dict[int, Dict[str, int]]) -> bool:
        """ Return True if player `pid` has at least `needed[resource]` of each resource """
        for resource, qty in needed.items():
            if all_resources[pid].get(resource, 0) < qty:
                return False
        return True

    def _exchange_resources(self, from_pid: int, to_pid: int,
                            offered: Dict[str, int], requested: Dict[str, int]):
        """
        Actually remove `offered` from `from_pid` and add to `to_pid`,
        and remove `requested` from `to_pid` and add to `from_pid`.
        """
        # For brevity. (In real code, always double-check you can do the exchange.)
        res = self.state.game_state["player_resources"]
        # from_pid loses offered
        for r, q in offered.items():
            res[from_pid][r] -= q
            res[to_pid][r]   += q
        # to_pid loses requested
        for r, q in requested.items():
            res[to_pid][r]   -= q
            res[from_pid][r] += q

    def _parse_resource_list(self, resource_str: str) -> Optional[Dict[str, int]]:
        """
        Example of parsing "2 Wheat, 1 Ore" into {"Wheat": 2, "Ore": 1}.
        Returns None on parse error.
        """
        # Split by commas or "and"
        items = re.split(r",\s*|\s+and\s+", resource_str, flags=re.IGNORECASE)
        parsed = {}
        for item in items:
            item = item.strip()
            if not item:
                continue
            # Expect format "<qty> <ResourceName>"
            match = re.match(r"(\d+)\s+(.+)", item)
            if not match:
                return None
            qty_str, rname = match.groups()
            qty = int(qty_str)
            # Attempt to standardize resource name
            rname = rname.strip().title()
            if rname not in self.resource_names:
                return None
            if qty <= 0:
                return None
            parsed[rname] = parsed.get(rname, 0) + qty
        return parsed if parsed else None

    def _offer_to_str(self, offered: Dict[str, int], requested: Dict[str, int]) -> str:
        """
        Return a quick string representation like "2 Wheat -> 3 Wood"
        """
        off_str = ", ".join(f"{q} {r}" for r, q in offered.items())
        req_str = ", ".join(f"{q} {r}" for r, q in requested.items())
        return f"{off_str} -> {req_str}"

    def _determine_winner(self):
        """
        Simple function if your game ends automatically at max_turns.
        You can define scoring logic, or declare a draw, or so on.
        """
        # Example: sum each player's total resource "market value"
        # Compare to find a winner
        game_state = self.state.game_state
        final_values = {}
        for pid in range(self.state.num_players):
            val = self._calculate_inventory_value(pid, game_state)
            final_values[pid] = val

        best_pid = max(final_values, key=final_values.get)
        best_val = final_values[best_pid]
        # Check if tie
        winners = [p for p, v in final_values.items() if v == best_val]
        if len(winners) == 1:
            reason=f"Player {best_pid} wins with a total value of {best_val}!"
            self.state.set_winners(player_ids=[best_pid], reason=reason)
        else:
            self.state.set_draw(reason=f"Tie among players {winners} with value {best_val}.")

    def _calculate_inventory_value(self, pid: int, game_state: Dict[str, Any]) -> int:
        """
        Summation of (count_of_resource * that_player's_value_for_resource).
        """
        total = 0
        resources = game_state["player_resources"][pid]
        values = game_state["player_values"][pid]
        for r in self.resource_names:
            total += resources[r] * values[r]
        return total
